-- dos32lib.e
-- dos version of Win32Lib, v.4
-- David Cuny
--  dcuny@lanset.com

constant VERSION = "0.4 pre-Alpha"
without warning
-- with trace

-- to do:
-- there is no need to redraw the whole desktop when another window gets
-- focus - I need to set up an getFocus and loseFocus event to handle it
-- instead of refreshing the entire desktop. at a minimum, that should
-- stop the awful flashing when a message box pops up.
-- isVisible takes parent's visiblilty into account, for my purposes.
-- perhaps I need to split it into another routine - isVisible gets the
-- attribute, isViewable says it can be drawn.

-- should probably have a "close modal" option or something for dialogs.
--   example: openDialog() and closeDialog()
-- move drag and double click into event loop
-- mouse needs to be hidden during graphics.

--$ Win32Lib Emulation
-- Dos32Lib is an emulation of the Win32Lib library. These routines are
-- behave more or less the same as their Win32Lib counterparts.


without warning
include graphics.e
include get.e
include mighty.e
include font.e  
include image.e
include wildcard.e
--include keyread.e
--include scan.e



global integer currentCtl

currentCtl = 0

-----------------------------------------------------------------------------
-- General Routines
-- The following are general support routines.


-----------------------------------------------------------------------------
function iif( integer flag, object o1, object o2 )
    -- Return a value based on a flag.
    -- This is resembles the
    -- \code iif()
    -- instruction from C, but evaluates both all parameters. As
    -- a result, it cannot be used as a "real"
    -- \code iff
    -- statement.
    
    if flag then
        return o1
    else
        return o2
    end if
end function

-----------------------------------------------------------------------------
function fix( atom a )
    -- convert an atom to an integer
    integer int

    -- get integer portion    
    int = floor( a )     
    
    -- round up or down?
    if a - int >= 0.5 then
        int = int + 1
    end if
    
    return int
    
end function


-----------------------------------------------------------------------------
function abs( atom a )
    -- return absolute value of a number

    if a > 0 then
        return a
    else
        return -a
    end if
    
end function



-----------------------------------------------------------------------------
function remove( object item, sequence list )
    -- Remove all instance of item from the list.
    
    integer at            
        
    while 1 do  
    
        -- look for item
        at = find( item, list )
    
        -- not more instances?
        if at = 0 then
            exit
        end if               
        
        -- remove instance
        list = list[1..at-1] & list[at+1..length(list)]
        
    end while  
        
    return list
    
end function


-----------------------------------------------------------------------------
function min( integer i1, integer i2 )
    if i1 < i2 then
        return i1
    else
        return i2
    end if
end function


-----------------------------------------------------------------------------
function max( integer i1, integer i2 )
    if i1 > i2 then
        return i1
    else
        return i2
    end if
end function


-----------------------------------------------------------------------------
global function or_all(sequence s)
-- or together all elements of a sequence
    atom result
    
    result = 0
    for i = 1 to length(s) do
        result = or_bits(result, s[i])
    end for
    return result
end function

-----------------------------------------------------------------------------
global function lo_word( atom a )
    -- return low word
    return and_bits( a, #FFFF )
end function

-----------------------------------------------------------------------------
global function hi_word( atom a )
    -- return high word
    return floor(a/power( 2, 16))
end function




integer abortHook
-----------------------------------------------------------------------------
procedure abortErr( sequence error, sequence args )
    -- print error and abort. 
    -- while loading, the routine uses a simple text message
    -- for run time, it pops up a dialog box.
    call_proc( abortHook, {error, args} )
end procedure


-----------------------------------------------------------------------------
procedure abortErr1( sequence error, sequence args )

    -- Simple abortErr behavior.
    -- This returns the video display to normal, prints a message,
    -- and aborts.
    
    integer result
    
    -- restore screen
    result = graphics_mode(-1)
    
    -- clear screen
    bk_color( BLACK )
    text_color( WHITE )
    clear_screen()
    
    -- print message
    printf( 1, error & "\n", args )
    
    -- abort
    abort(0)
    
end procedure

abortHook = routine_id("abortErr1")



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Widget Class Attributes


-- control classes
-- Win32Lib    
-- not all controls are supported yet
global constant
    Window          = 1,
    PushButton      = 2,
    DefPushButton   = 3,
    CheckBox        = 4,    
    TriCheckBox     = 5,        -- not implemented
    Radio           = 6,
    Group           = 7,        -- not implemented
    List            = 8,        -- not implemented
    SortedList      = 9,        -- not implemented
    Combo           = 10,       -- not implemented
    SortedCombo     = 11,       -- not implemented
    SimpleCombo     = 12,       -- not implemented
    DropDownList    = 13,       -- not implemented
    EditText        = 14,
    MleText         = 15,       -- not implemented
    LText           = 16,
    RText           = 17,
    CText           = 18,       -- not implemented
    HScroll         = 19,
    VScroll         = 20,
    Icon            = 21,
    Bitmap          = 22,
    Menu            = 23,       -- NEW! 0.13
    MenuItem        = 24        -- NEW! 0.13
    
    
constant
    ControlCount    = MenuItem

-- these controls can't receive events
constant StaticControls = { LText, RText, Icon, Bitmap }


-- class actions
global sequence
    classShow,      -- display a control
    classClick,     -- handle a click event
    classKey        -- handle a key press

    classShow       = repeat( -1, ControlCount )
    classClick      = repeat( -1, ControlCount )
    classKey        = repeat( -1, ControlCount )
    

constant    
    Class       = 1,
    Text        = 2,
    Parent      = 3,
    Size        = 4,
    Flags       = 5,
    State       = 6,        -- active/inactive/checked
    Data        = 7,        -- data
    MenuList    = 8,        -- true if menu exists
    Range       = 9,        -- {min, max, small change, large change}
    PenPos      = 10,       -- pen position
    PenColor    = 11,       -- pen color
    BackColor   = 12,       -- background color of control
    FontAttrib  = 13,       -- font attribute
    FontSize    = 14,       -- font size
    Font        = 15,       -- font
    MousePtr    = 19,       -- mouse pointer associated with control
    Focus       = 20,       -- id of control with focus/position of focus
    TabStops    = 21,       -- controls that can be tabbed to
    Anchor      = 22,       -- offset for list, sle
    Modal       = 23,       -- true if window is modal
    Bmp         = 24,       -- bitmap image associated with control
    PointerAt   = 25,       -- position of pointer/thumb
    
    MySize  = PointerAt


-- control states
constant                                        
    IsVisible       = 1,
    IsEnabled       = 2,    -- control is active
    IsChecked       = 4,    -- control is checked
    IsPressed       = 8     -- control is pressed by mouse

-- button states
constant
    ActiveUnchecked     = 1,        -- is "armed" is better than "pressed?"
    ActiveChecked       = 2,
    PressedUnchecked    = 3,
    PressedChecked      = 4,
    InactiveUnchecked   = 5,
    InactiveChecked     = 6,
    
-- scroll bar value indexes for Range
    ScrollMin           = 1,
    ScrollMax           = 2,
    ScrollSmall         = 3,
    ScrollLarge         = 4

-- mouse action
global constant
    MOUSE_MOVE          = MOVE

sequence my

-- window styles
global constant
    WS_OVERLAPPED   = #0,
    WS_POPUP        = #80000000,
    WS_CHILD        = #40000000,
    WS_MINIMIZE     = #20000000,
    WS_VISIBLE      = #10000000,
    WS_DISABLED     = #8000000,
    WS_CLIPSIBLINGS = #4000000,
    WS_CLIPCHILDREN = #2000000,
    WS_MAXIMIZE     = #1000000,
    WS_CAPTION      = #C00000,      --  WS_BORDER Or WS_DLGFRAME
    WS_BORDER       = #00800000,    -- creates border on window
    WS_DLGFRAME     = #400000,
    WS_HSCROLL      = #00100000,    -- horizontal scroll bar
    WS_VSCROLL      = #00200000,    -- vertical scroll bar
    WS_SYSMENU      = #80000,
    WS_THICKFRAME   = #40000,
    WS_GROUP        = #20000,
    WS_TABSTOP      = #10000,   -- use tab stop

    WS_MINIMIZEBOX = #20000,
    WS_MAXIMIZEBOX = #10000,

    WS_TILED = WS_OVERLAPPED,
    WS_ICONIC = WS_MINIMIZE,
    WS_SIZEBOX = WS_THICKFRAME,
    WS_OVERLAPPEDWINDOW = or_all({  WS_OVERLAPPED, 
                                    WS_CAPTION, 
                                    WS_SYSMENU, 
                                    WS_THICKFRAME, 
                                    WS_MINIMIZEBOX, 
                                    WS_MAXIMIZEBOX}),
    
    
    WS_TILEDWINDOW = WS_OVERLAPPEDWINDOW,

--
--   Common Window Styles
--  /


    WS_POPUPWINDOW = or_all({WS_POPUP, WS_BORDER, WS_SYSMENU}),

    WS_CHILDWINDOW = (WS_CHILD)
    


-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--$ Color Scheme

-----------------------------------------------------------------------------
-- set up video mode 
integer res
sequence vAttrib, cmd

-- default resolution
    res = 18

-- read the command line
cmd = command_line()
if length(cmd) = 3 then
    -- convert it
    cmd = value( cmd[3] )
    if cmd[1] = GET_SUCCESS then
        -- use abs() to strip of '-'
        res = abs(cmd[2])
    end if
end if


if graphics_mode(res) then
    abortErr( "Unable to set graphics mode %d", {res} )
end if

-- screen size    
global integer 
    xPixels,            -- screen x pixels
    yPixels             -- screen y pixels

    xPixels = 0
    yPixels = 0

    
-- get the attribute
vAttrib = video_config()
xPixels = vAttrib[ VC_XPIXELS ]
yPixels = vAttrib[ VC_YPIXELS ]



-----------------------------------------------------------------------------
procedure setPal( integer color, sequence code )
    --$ Color Scheme
    -- Set an rgb color.
    object ignore       
    ignore = palette( color, floor( code/4) )
end procedure


-----------------------------------------------------------------------------
-- Win32 System Colors  
global constant
    COLOR_SCROLLBAR         = 1,    -- scrollbar
    COLOR_BACKGROUND        = 2,    -- desktop
    COLOR_ACTIVECAPTION     = 3,    -- active window caption
    COLOR_INACTIVECAPTION   = 4,    -- inactive window caption
    COLOR_MENU              = 5,    -- menu background
    COLOR_WINDOW            = 6,    -- window background
    COLOR_WINDOWFRAME       = 7,    -- window frame
    COLOR_MENUTEXT          = 8,    -- menu text
    COLOR_WINDOWTEXT        = 9,    -- window text
    COLOR_CAPTIONTEXT       = 10,   -- in caption, size box and scroll arrow
    COLOR_ACTIVEBORDER      = 11,   -- active border
    COLOR_INACTIVEBORDER    = 12,   -- inactive border
    COLOR_APPWORKSPACE      = 13,   -- mdi background color
    COLOR_HIGHLIGHT         = 14,   -- items selected in control
    COLOR_HIGHLIGHTTEXT     = 15,   -- text of items selected in control
    COLOR_BTNFACE           = 16,   -- NT: face shading on buttons
    COLOR_BTNSHADOW         = 17,   -- edge shading on push buttons
    COLOR_GRAYTEXT          = 18,   -- disabled text
    COLOR_BTNTEXT           = 19,   -- button text
    COLOR_INACTIVECAPTIONTEXT = 20, -- text in inactive caption
    COLOR_BTNHIGHLIGHT      = 21    -- button highlight color

sequence sysColor
    sysColor = repeat( 0, 21 )    

    -- window background
    sysColor[COLOR_BACKGROUND]          = BRIGHT_BLUE --WHITE
    
    -- window title bar    
    sysColor[COLOR_ACTIVECAPTION]       = BLUE
    sysColor[COLOR_CAPTIONTEXT]         = BRIGHT_WHITE
    sysColor[COLOR_INACTIVECAPTION]     = GRAY
    sysColor[COLOR_INACTIVECAPTIONTEXT] = WHITE

    -- the window
    sysColor[COLOR_WINDOW]              = WHITE
    sysColor[COLOR_WINDOWFRAME]         = WHITE
    sysColor[COLOR_WINDOWTEXT]          = BLACK
    sysColor[COLOR_ACTIVEBORDER]        = BLUE
    sysColor[COLOR_INACTIVEBORDER]      = GRAY
    sysColor[COLOR_APPWORKSPACE]        = BRIGHT_WHITE

    -- the menu bar
    sysColor[COLOR_MENU]                = WHITE --BRIGHT_BLUE
    sysColor[COLOR_MENUTEXT]            = BLACK

    -- the controls    
    sysColor[COLOR_SCROLLBAR]           = BRIGHT_WHITE
    sysColor[COLOR_HIGHLIGHT]           = BLUE
    sysColor[COLOR_HIGHLIGHTTEXT]       = BRIGHT_WHITE
    sysColor[COLOR_BTNFACE]             = WHITE
    sysColor[COLOR_BTNSHADOW]           = GRAY
    sysColor[COLOR_GRAYTEXT]            = GRAY
    sysColor[COLOR_BTNTEXT]             = BLACK
    sysColor[COLOR_BTNHIGHLIGHT]        = BRIGHT_WHITE
    

-- adjust the colors
setPal( WHITE,         {204, 204, 204} )
setPal( GRAY,          {102, 102, 102} )
setPal( BLUE,          { 51,  51, 153} )
setPal( BRIGHT_BLUE,          { 99,  99, 206} )         
setPal( BRIGHT_BLUE,   {153, 153, 204} )         
setPal( BROWN,         {204, 153, 102} )
setPal( BRIGHT_CYAN,   {156, 206, 206} )
setPal( CYAN,          {156, 206, 206} )
setPal( MAGENTA,       { 96,   0,  96} )


-----------------------------------------------------------------------------
-- window metrics
integer
    theBorderSize,      -- thickness of border
    theBorderEdge,      -- which border edge is selected
    theGripSize,        -- thickness of grip
    theTitleSize,       -- height of title bar
    theMenuSize,        -- height of menu
    theMousePointer     -- current mouse pointer
    
-- sizes
    theBorderSize      = 5
    theGripSize        = 15
    theTitleSize       = 18
    theMenuSize        = 18
    theMousePointer = IDC_ARROW
    

constant
    X1  = 1,
    Y1  = 2,
    X2  = 3,
    Y2  = 4

--    IDC_APPSTARTING = 13, -- Standard arrow and small hourglass
--    IDC_ARROW = 1,        --  Standard arrow
--    IDC_CROSS = 3,        -- Crosshair
--    IDC_IBEAM = 4,        -- Text I-beam
--    IDC_ICON = 5,         -- Empty icon
--    IDC_NO = 6,           -- Slashed circle
--    IDC_SIZE = 7,         -- Four-pointed arrow
--    IDC_SIZEALL = 7,      -- Same as MC_SIZE
--    IDC_SIZENESW = 8,     -- Double-pointed arrow pointing northeast and southwest
--    IDC_SIZENS = 9,       -- Double-pointed arrow pointing north and south
--    IDC_SIZENWSE  = 10,   -- Double-pointed arrow pointing northwest and southeast
--    IDC_SIZEWE = 11,      -- Double-pointed arrow pointing west and east
--    IDC_UPARROW = 12,     -- Vertical arrow
--    IDC_WAIT = 2          -- Hourglass

-- compass directions
constant    
    North   = 1,
    South   = 2,
    East    = 4,
    West    = 8



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--$ Rendering Code
-- These routines enhance the built-in Euphoria routines by adding
-- clipping code.

sequence 
    clipWin     -- clipping window
    clipWin = { 0, 0, xPixels, yPixels }
    

-- bit masks of clip results
constant
    LeftEdge    = 1,
    RightEdge   = 2,
    TopEdge     = 4,
    BottomEdge  = 8



-----------------------------------------------------------------------------
function getClipCode( integer x, integer y )
    -- Support function for clipLine().
    -- This returns bit mask indicating where the point lies,
    -- relative to the clipping window
    integer code

    -- default value
    code = 0
    
    -- check for intersection with top/bottom
    if y < clipWin[Y1] then
        code = TopEdge
        
    elsif y > clipWin[Y2] then
        code = BottomEdge
        
    end if

    -- check for intersection with left/right
    if x < clipWin[X1] then
        code = code+LeftEdge
        
    elsif x > clipWin[X2] then
        code = code+RightEdge
        
    end if

    return code
        
end function


-----------------------------------------------------------------------------
procedure clipLine( integer color, integer x1, integer y1, 
                        integer x2, integer y2 )
    -- Draws a clipped line.    
    -- This routine takes a color and a set of points in the form:
    -- \p {x1, y2, x2, y2}
    -- and draws a clipped line.
    
    integer 
    code1,      -- clip code for {x1,y1} 
    code2,      -- clip code for {x2,y2}
    accept,     -- if true, draws the line
    code,       -- current clip code being examined
    x, y        -- intersection for current clip code
        
    -- get clip codes for end points
    code1 = getClipCode( x1, y1 )
    code2 = getClipCode( x2, y2 )

    -- default is not to accept
    accept = 0
    
    -- clip one, or both points
    while 1 do                             

        -- line lies entirely inside of the clipping area?
        if or_bits( code1, code2 ) = 0 then             
            -- set flags
            accept = 1
            exit
        end if
        
        -- line lies entirely outside of bounds?
        if and_bits( code1, code2 ) != 0 then
            -- trivial reject
            accept = 0
            exit
        end if
    
        -- pick one of the offending points
        if code1 != 0 then
            code = code1
        else
            code = code2
        end if

        -- clip against left or right?
        if and_bits( code, LeftEdge ) then
            -- clip against left edge
            y = fix( y1+(y2-y1)*(clipWin[X1]-x1) / (x2-x1) )
            x = clipWin[X1]
        
        elsif and_bits( code, RightEdge ) then        
            -- clip against bottom edge  
            y = fix( y1+(y2-y1)*(clipWin[X2]-x1) / (x2-x1) )
            x = clipWin[X2]
            
        -- clip against top or bottom?
        elsif and_bits( code, TopEdge ) then
            -- clip against top edge
            x = fix( x1+(x2-x1)*(clipWin[Y1]-y1) / (y2-y1) )
            y = clipWin[Y1]
        
        elsif and_bits( code, BottomEdge ) then
            -- clip against bottom edge
            x = fix(x1+(x2-x1)*(clipWin[Y2]-y1) / (y2-y1) )
            y = clipWin[Y2]
            
        end if


        -- replace the point in the line
        if code = code1 then
            -- clipped first point
            x1 = x
            y1 = y   
            
            -- recalculate code
            code1 = getClipCode( x1, y1 )
            
        else
            -- clipped second point
            x2 = x
            y2 = y
            
            -- recalculate code
            code2 = getClipCode( x2, y2 )
            
        end if
        
    end while
        
    -- if accepted, draw line
    if accept then  
        mouse_pointer( 0 )
        draw_line( color, { {x1,y1}, {x2,y2} } )
        mouse_pointer( 1 )
    end if
    
end procedure


-----------------------------------------------------------------------------
procedure clipImage( sequence point, sequence image )
    -- Draws a clipped bitmap.
    -- This routine takes a point and a bitmap, clips it to the
    -- current clipping window, and displays it.
    
    integer x1, y1, x2, y2, diff, start, finish

    -- mangle point; hack
    point = point - 1
    
    -- get size
    x1 = point[X1]
    y1 = point[Y1]
    x2 = x1 + length(image[1]) - 1
    y2 = y1 + length(image) - 1

    -- off the window?    
    if x2 < clipWin[X1]
    or x1 > clipWin[X2]
    or y2 < clipWin[Y1]
    or y1 > clipWin[Y2] then
        return
    end if    

    -- perform "cheap" clipping first
    
    -- off top edge?
    if y1 < clipWin[Y1] then
        -- clip
        diff = clipWin[Y1] - y1
        image = image[diff..length(image)]
        -- new values
        y1 = clipWin[Y1]
        y2 = y2 - diff        
    end if

    -- off bottom?
    if y2 > clipWin[Y2] then
        -- clip
        diff = y2 - clipWin[Y2]
        image = image[1..length(image)-diff]
    end if


    -- off right or left edge? these are put into one test 
    -- so I go through the loop only once
    if x1 < clipWin[X1]
    or x2 > clipWin[X2] then
        
        -- clip left? 
        if x1 < clipWin[X1] then
            -- difference
            start = clipWin[X1] - x1
            x1 = x1 + start
        else
            -- no change
            start = 1
        end if
        
        -- clip right?
        finish = length( image[1] )
        if x2 > clipWin[X2] then        
            -- difference
            finish = finish - (x2 - clipWin[X2])
        end if

        -- clip        
        for i = 1 to length( image ) do
            image[i] = image[i][start..finish]
        end for       
    end if

    -- hide the mouse
    mouse_pointer( 0 )

    -- display
    display_image( { x1, y1 }, image )

    -- show the mouse
    mouse_pointer( 1 )
    
    
end procedure





-----------------------------------------------------------------------------
procedure clipXorLine( integer color, integer x1, integer y1, 
                                        integer x2, integer y2 )
    -- Draw a clipped XOR line.
    -- This <i>only</i> draws horizontal or vertical lines.
    -- Drawing an XOR line twice erases it.
    
    integer tmp, len
    sequence s

    -- hide the mouse
    mouse_pointer( 0 )

    -- horizontal, or vertical?
    if x1 = x2 then        
        -- horizontal
        if y1 > y2 then
            tmp = y1
            y2 = y1
            y1 = tmp
        end if      
        
        for y = y1 to y2 do
            pixel(xor_bits(get_pixel({x1,y,1}),color),{x1,y})                
        end for
        
    else
        -- vertical
        if x2 < x1 then
            -- swap
            tmp = x1
            x1 = x2
            x2 = x1
        end if
        
        -- length
        len = x2-x1+1

        -- line to xor against
        s = repeat( color, len )
        
        -- get the line, xor it, and draw it
        pixel(xor_bits(get_pixel({x1,y1,len}),s),{x1,y1})
        
    end if

    -- show the mouse
    mouse_pointer( 1 )

    
end procedure



-----------------------------------------------------------------------------
procedure clipXorRect( integer color, sequence r )

    -- draw an XOR rectangle.
    -- drawing an XOR rectangle twice erases it.

    integer x1, y1, x2, y2

    -- get points    
    x1 = r[X1]
    y1 = r[Y1]
    x2 = r[X2]
    y2 = r[Y2]

    -- draw a rectangle
    -- need to avoid overlapping, or corners will disappear
    clipXorLine( color, x1, y1, x2, y1 )
    clipXorLine( color, x2, y1+1, x2, y2-1 )
    clipXorLine( color, x1, y2, x2, y2 )
    clipXorLine( color, x1, y1+1, x1, y2-1 )
    
end procedure


-----------------------------------------------------------------------------
procedure clipPixel( object o, integer x1, integer y1 )

    -- need to add clipping code!
    integer x2, p1, p2, diff
    
    -- single point?            
    if integer( o ) then
        -- out of clip window?
        if x1 < clipWin[X1]
        or x1 > clipWin[X2]
        or y1 < clipWin[Y1]
        or y1 > clipWin[Y2] then
            -- do nothing
        else
            mouse_pointer( 0 )            
            pixel( o, {x1, y1} )
            mouse_pointer( 1 )
        end if          
        
    -- seqence of pixels
    else            

        -- end position
        x2 = x1+length( o )-1
        
        -- in clip area?
        if x2 < clipWin[X1]
        or x1 > clipWin[X2]
        or y1 < clipWin[Y1]
        or y1 > clipWin[Y2] then
            -- do nothing
        else
            -- clip ends
            if x1 < clipWin[X1] then
                p1 = clipWin[X1] - x1
            else
                p1 = 1
            end if
            
            if x2 > clipWin[X2] then
                p2 = length(o) - (x2 - clipWin[X2])
            else
                p2 = length(o)
            end if

            -- display                        
            mouse_pointer( 0 )            
            pixel( o[p1..p2], {x1, y1} )
            mouse_pointer( 1 )
                            
        end if
    end if
                                
end procedure


-----------------------------------------------------------------------------
procedure clipFocusRect( integer fillColor, integer x1, integer y1,
                                integer x2, integer y2 )

    -- draw a focus rectangle
    integer long
    sequence pix

    long = x2 - x1    
    pix = repeat( fillColor, long )
    for i = 1 to long by 2 do
        pix[i] = BLACK
    end for
    clipPixel( pix, x1, y1 )
    clipPixel( pix, x1, y2 )

    -- vertical lines
    for i = y1 to y2 by 2 do
        clipPixel( BLACK, x1, i )
        clipPixel( BLACK, x2, i )
    end for

                                
end procedure


-----------------------------------------------------------------------------
procedure clipRect( integer color, integer filled, integer x1, integer y1,
                                integer x2, integer y2 )
                                
    -- Displays a clipped box, optionally filled.
    
    -- outside of clip box?
    if x1 > clipWin[X2]
    or x2 < clipWin[X1]
    or y1 > clipWin[Y2]
    or y2 < clipWin[Y1] then
        -- nothing to draw
        return
    end if

    if filled then
        -- clip and use polygon. 
        -- faster than putting clipLine in a loop
        
        if x1 < clipWin[X1] then x1 = clipWin[X1] end if
        if x2 > clipWin[X2] then x2 = clipWin[X2] end if
        if y1 < clipWin[Y1] then y1 = clipWin[Y1] end if
        if y2 > clipWin[Y2] then y2 = clipWin[Y2] end if

        mouse_pointer( 0 )

        -- draw a filled polygon    
        polygon( color, 1, { { x1, y1 },
                             { x2, y1 },
                             { x2, y2 },
                             { x1, y2 } } )                                
                             
        mouse_pointer( 1 )
        
        
    else                         
        -- call clipLine to draw a box
        clipLine( color, x1, y1, x2, y1 )
        clipLine( color, x2, y1, x2, y2 )
        clipLine( color, x2, y2, x1, y2 )
        clipLine( color, x1, y2, x1, y1 )
    end if
    
end procedure


-----------------------------------------------------------------------------
procedure clipFramedRect( integer color, integer x1, integer y1,
                                integer x2, integer y2 )

    -- draw a rectangle with a black border
    clipRect( color, 1, x1, y1, x2, y2 )
    clipRect( BLACK, 0, x1, y1, x2, y2 )
    
                                
end procedure                                


-----------------------------------------------------------------------------
procedure clipShadowedRect( integer color, integer x1, integer y1,
                                integer x2, integer y2 )

    -- draw a rectangle with a black border and shadows
    -- shadow is added past size of rectangle
    
    clipRect( color, 1, x1, y1, x2, y2 )
    clipRect( BLACK, 0, x1, y1, x2, y2 )               
    -- shadow
    clipLine( BLACK, x2+1, y1+1, x2+1, y2+1 )
    clipLine( BLACK, x1+1, y2+1, x2+1, y2+1 )
    
                                
end procedure                                

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Routines for Drawing into Bitmaps
-- these need to eventually be folded into screen drawing tools

-----------------------------------------------------------------------------
function bitmapPixel( sequence bmp, integer color, integer x, integer y )

    -- draws a pixel on a bitmap, with clipping
    -- needs to eventually be expanded to handle sequence of pixels

    if y > 0 and y <= length( bmp ) then
        if x > 0 and x <= length( bmp[1] ) then
            bmp[y][x] = color
        end if
    end if
    return bmp
end function  


-----------------------------------------------------------------------------
function bitmapLine( sequence bmp, integer color, integer x1, integer y1,
                            integer x2, integer y2 )
                            
    -- draws a line on a bitmap, with clipping

    -- integer direction
    integer direction, maxX, maxY
    atom deltaX, deltaY, x, y

    -- get bounds
    maxX = length(bmp[1])
    maxY = length(bmp)

    -- draw a line in a bitmap
    if x1 = x2 then
       -- vertical line 
       
        -- swap order?
        if y2 < y1 then
            -- swap
            y  = y1
            y1 = y2
            y2 = y1
        end if
       
        -- in range?
        if  y1 <= maxY
        and y2 > 0 then

            -- clip
            y1 = iif( y1 < 1, 1, y1 )
            y2 = iif( y2 > maxY, maxY, y2 )

            -- draw            
            for i = y1 to y2 do
                bmp[i][x1] = color
            end for
        
        end if
       
    
    elsif y1 = y2 then
        -- horizontal line

        -- swap order?
        if x2 < x1 then
            -- swap
            x  = x1
            x1 = x2
            x2 = x1
        end if


        -- in range?
        if  x2 > 0
        and x1 <= maxX then
                  
            -- clip
            x1 = iif( x1 < 1, 1, x1 )
            x2 = iif( x2 > maxX, maxX, x2 )
            
            -- draw
            bmp[y1][x1..x2] = color
            
        end if
    
    else
        -- diagonal
        deltaX = x2 - x1
        deltaY = y2 - y1

        -- move on x or y axis?        
        if abs( deltaX ) > abs( deltaY ) then
            
            -- positive or negative slope?
            direction = iif( x1>x2, 1, -1 )
            
            -- start point
            y = y1

            -- draw the line along x axis
            for i = x1 to x2 by direction do
                bmp = bitmapPixel( bmp, color, i, floor(y) )
                y = y + deltaY
            end for

        else
            
            -- positive or negative slope?
            direction = iif( y1 > y2, 1, -1 )
            
            -- start point
            x = x1

            -- draw the line            
            for i = y1 to y2 by direction do
                bmp = bitmapPixel( bmp, color, i, floor(y) )
                x1 = x + deltaX
            end for

            
        end if
        
    end if 
    
    return bmp
    
end function  


-----------------------------------------------------------------------------
function bitmapBox( sequence bmp, integer color, integer fill, 
                            integer x1, integer y1,
                            integer x2, integer y2 )

    integer maxX, maxY

    -- draw a box in a bitmap
    if fill then               
    
        -- get the max points
        maxX = length( bmp[1] )
        maxY = length( bmp )
    
        -- inside the bounds?
        if  x1 <= maxX and x2 < 1
        and y1 <= maxY and y2 < 1 then
        
            -- clip the points
            x1 = iif( x1 < 1, 0, x1 )
            y1 = iif( y1 < 1, 0, y1 )
            x2 = iif( x2 > maxX, x2, maxX )
            y2 = iif( y2 > maxY, y2, maxY )

            -- fill        
            for y = y1 to y2 do
                bmp[y][x1..x2] = color
            end for
            
        end if
        
    else
        -- draw lines
        -- ordered for speed
        bmp = bitmapLine( bmp, color, x1, y1, x2, y1 )
        bmp = bitmapLine( bmp, color, x1, y2, x2, y2 )
        bmp = bitmapLine( bmp, color, x1, y1, x1, y2 )
        bmp = bitmapLine( bmp, color, x2, y1, x2, y2 )

    end if

    
    return bmp

end function

-----------------------------------------------------------------------------
function saveImage( integer x1, integer y1, integer x2, integer y2 )
    -- "safe" version of save_image; handles negative indexes
    
    -- clip to edge of screen
    return save_image( {max(x1,0),max(y1,0)}, {max(x2,0),max(y2,0)} )


end function    
-----------------------------------------------------------------------------
procedure restoreImage( integer x1, integer y1, sequence s )

    -- "safe" version of display_image; handles negative indexes
    
    -- clip to edge of screen
    if x1 < 0 then
        x1 = 0
    end if
    
    if y1 < 0 then
        y1 = 0
    end if
    
    display_image( {max(x1,0),max(y1,0)}, s )

end procedure


-----------------------------------------------------------------------------
constant
    LeftJustify     = 1,
    RightJustify    = 2,
    CenterJustify   = 4,
    TopJustify      = 8,
    BottomJustify   = 16,
    MiddleJustify   = 32
    
-----------------------------------------------------------------------------
function justifyAt( integer x1, integer y1, integer x2, integer y2, 
                integer flags, sequence image )

    -- return position to place image at

    integer tall, wide, atX, atY

    -- metrics of image    
    tall    = length(image)
    wide    = length( image[1] )

    -- horizontal    
    if and_bits( flags, RightJustify ) then
        -- right justify
        atX = x2 - wide
    elsif and_bits( flags, CenterJustify ) then
        -- center justify
        atX = floor(x1 + (( x2-x1 )/2 ) - (wide/2))
    else
        -- default, left justify
        atX = x1
    end if

    -- vertical
    if and_bits( flags, TopJustify ) then
        -- top justify
        atY = y1
    elsif and_bits( flags, BottomJustify ) then
        -- bottom justify
        atY = y2 - tall
    else
        -- center justify
        atY = floor(y1 + (( y2-y1 )/2 ) - (tall/2))+1
    end if

    -- draw it
    return {atX, atY}
    
end function


-----------------------------------------------------------------------------
procedure clipText( integer x, integer y, sequence text, 
                    integer fore, integer back, integer attrib,
                    integer focus )

    -- draw text, with optional focus rectangle
    
    integer x2, y2
    sequence image
    
    image = build_text_image( text, fore, back, attrib )

    -- display
    clipImage( {x, y}, image )
    
    -- focus?
    if focus then
        -- extent
        x2 = x + length( image[1] ) - 1
        y2 = y + length( image ) - 1

        -- focus        
        clipFocusRect( back, x, y, x2, y2 )
        
    end if

end procedure


-----------------------------------------------------------------------------
function buildImage( sequence s, integer fColor, integer bColor )
    -- create an image of the controls

    -- replace colors
    for i = 1 to length( s ) do
        for j = 1 to length( s[i] ) do
            if      s[i][j] = ' ' then s[i][j] = bColor
            elsif   s[i][j] = 'x' then s[i][j] = fColor
            elsif   s[i][j] = 'X' then s[i][j] = GRAY
            elsif   s[i][j] = '.' then s[i][j] = BRIGHT_WHITE
            elsif   s[i][j] = '/' then s[i][j] = WHITE
            else    s[i][j] = fColor
            end if
        end for
    end for

    return s

end function




-----------------------------------------------------------------------------
-- GENERIC CONTROLS
--

-----------------------------------------------------------------------------
function isVisible( integer self )
    -- Win32
    -- returns true if control is visible    
    integer parent

    -- get parent of control
    parent = my[self][Parent]
    if parent then
        -- parent and control must be visible
        return  and_bits( my[parent][State], IsVisible ) and
                and_bits( my[self][State], IsVisible )
    else
        -- just window
        return and_bits( my[self][State], IsVisible )
    end if
    
    
end function

-----------------------------------------------------------------------------
global function isEnabled( integer self )
    -- Win32
    return and_bits( my[self][State], IsEnabled )
end function

-----------------------------------------------------------------------------
global function isChecked( integer self )
    -- Win32    
    return and_bits( my[self][State], IsChecked )
end function

-----------------------------------------------------------------------------
function isPressed( integer self )
    -- Win32    
    return and_bits( my[self][State], IsPressed )
end function


-----------------------------------------------------------------------------
procedure showControl( integer self )

    -- draw a control    
    integer class, parent, procId

    -- don't display if parent is not visible
    parent = my[self][Parent] 
    if parent > 0 then
        if not isVisible( parent ) then
            return
        end if        
    end if

    -- don't draw if is invisible
    if not isVisible( self ) then
        return
    end if
    
    
    class = my[self][Class]    
    procId = classShow[class]
    if procId > -1 then
        -- hide the mouse
        mouse_pointer(0)
        
        -- call the proc
        call_proc( classShow[class], {self} )        
        
        -- restore the mouse
        mouse_pointer(1)
    end if
    
end procedure    












-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Fonts

-----------------------------------------------------------------------------
-- currently, only 10 point fonts are supported
constant FontNames = {
    "Arial",                "Arial10.f",
    "Ariel",                "Arial10.f",
    "Bookman Old Style",    "Book10.f",
    "Comic Sans MS",        "Comic10.f",
    "Courier",              "Cour10.f",
    "Courier New",          "Cour10.f",
    "Fixedsys",             "FixSys10.f",
    "MS Sans Serif",        "Sans10.f",
    "Small Fonts",          "Small07",
    "System",               "System10.f",
    "Tahoma",               "Tahoma10.f",
    "Times New Roman",      "Times10.f",
    "Verdana",              "Verdan10.f" }




-----------------------------------------------------------------------------
-- load the fonts
integer
    SysFont, FixedFont

-- system font
SysFont = fload( "verdan08.f" )
FixedFont = fload("fixsys08.f")

-----------------------------------------------------------------------------

-- user defined actions: 1 per control
-- Win32Lib
global sequence
-- NEW! 0.4 renamed from onLoad to onOpen
    onOpen,
    onPaint,
    onClick,
    onMouse,
    onRightClick,
    onScroll,
    onKeyPress,
-- NEW! 0.4 added
    onKeyDown,
    onGetFocus,
    onLostFocus,
    onTimer

    onOpen  = {}    
    onPaint = {}    
    onClick = {}
    onMouse = {}
    onRightClick = {}
    onScroll = {}  
-- NEW! 0.4 added
    onKeyDown = {}
    onKeyPress = {}
    onGetFocus = {}
    onLostFocus = {}
    onTimer = {}
    
    
    
-- timers  
sequence timers
timers = {}         -- holds timer list

constant
    TimerWin    = 1,
    TimerId     = 2,
    TimerTicks  = 3,
    TimerStart  = 4

-- window stack
sequence
    windowStack     -- stack of open windows
    windowStack = {}
    
-- mouse controls
constant
    MOUSE_EVENT     = 1,    -- used by get_mouse
    MOUSE_BUTTONS   = 1,    -- used by poll_mouse
    MOUSE_X         = 2,
    MOUSE_Y         = 3,
    MOUSE_LEFT      = 1,
    MOUSE_RIGHT     = 2,
    MOUSE_MIDDLE    = 3

-- event loop flag
integer exitEventLoop
    exitEventLoop = 0

-----------------------------------------------------------------------------
integer
    mainWindow
    
    mainWindow = 0      -- window eventLoop() was entered with

global constant
    Default = -1

-- colors
-- NEW! 0.4 added more colors
global constant              
    Black           = BLACK,
    Blue            = BLUE,
    Green           = GREEN,
    Magenta         = MAGENTA,
    Red             = RED,
    Cyan            = CYAN,
    Brown           = BROWN,
    White           = WHITE,
    Gray            = GRAY,
    BrightBlue      = BRIGHT_BLUE,
    BrightGreen     = BRIGHT_GREEN,
    BrightMagenta   = BRIGHT_MAGENTA,
    BrightRed       = BRIGHT_RED,
    BrightCyan      = BRIGHT_CYAN,
    Yellow          = YELLOW,
    BrightWhite     = BRIGHT_WHITE
    
    
-- font attributes
global constant
    Bold            = 1,
    Italic          = 2,
    Underline       = 4,
    Strikeout       = 8


-- holds attributes
my = {}


-----------------------------------------------------------------------------
-- keys
-- NEW! 0.4 increased number of keys defined
global constant    
                
    VK_BACK = 8, 
    VK_TAB = 9, 
    VK_RETURN = 13, 
    VK_ESCAPE = 27,
    VK_SPACE = 32,
    VK_SHIFT_TAB = 271,
    VK_HELP = 315, 

    VK_F1 = 315, 
    VK_F2 = 316, 
    VK_F3 = 317, 
    VK_F4 = 318, 
    VK_F5 = 319,    
    VK_F6 = 320, 
    VK_F7 = 321, 
    VK_F8 = 322, 
    VK_F9 = 323, 
    VK_F10 = 324,
    VK_F11 = 389, 
    VK_F12 = 390,

    VK_HOME = 327, 
    VK_UP = 328, 
    VK_PRIOR = 329, 
    VK_LEFT = 331,
    VK_RIGHT = 333, 
    VK_END = 335, 
    VK_DOWN = 336, 
    VK_NEXT = 337, 
    VK_INSERT = 338, 
    VK_DELETE = 339, 

    -- my additions
    VK_PAGEUP   = VK_PRIOR,
    VK_PAGEDOWN = VK_NEXT
    
-- non-Win32
-- NEW! 0.4
constant
    SHIFT_TAB = 271,
    
    CTRL_LEFT = 371, 
    CTRL_RIGHT = 372, 
    CTRL_END = 373,
    CTRL_HOME = 375, 
    CTRL_PAGE_UP = 388, 
    CTRL_PAGE_DOWN = 374,
    CTRL_UP = 397, 
    CTRL_DOWN = 401, 
    CTRL_INSERT = 402,
    CTRL_TAB = 404, 
    CTRL_ENTER = 10,

    ALT_TAB = 421

    

    
-----------------------------------------------------------------------------
integer
    winX1,      -- position of window's X postition
    winY1,      -- position of window's Y position
    winX2,
    winY2

-----------------------------------------------------------------------------
function hasMenu( integer self )
    -- returns count of menu items
    -- ex:  ? menuItems( MyWindow )
    --      --> 4
    return length( my[self][MenuList] )
end function


-----------------------------------------------------------------------------
function getClientSize( integer self )            

    -- returns size of client area
    sequence size

    -- get the window size
    size = my[self][Size][X2..Y2]

    -- subtract some pixels
    size = size - { 3, 3 }

    -- subtract the title bar
    size[Y1] = size[Y1] - (theTitleSize)

    -- add menu?       
    if hasMenu( self ) then
        size[Y1] = size[Y1] - (theMenuSize)
    end if
    
    return size

end function


-----------------------------------------------------------------------------
global function getClientOrg( integer self )

    -- Win32 API
    -- returns origin of client area
    sequence org

    -- get the coordinates
    org = my[self][Size][X1..Y1]

    -- add border, plus pixels
    org[X1..Y1] = org[X1..Y1] + { 2, 1 }

    -- add title bar
    org[Y1] = org[Y1] + theTitleSize

    -- add menu?       
    if hasMenu( self ) then
        org[Y1] = org[Y1] + theMenuSize
    else
        -- seperator
        org[Y1] = org[Y1] + 1
    end if
    
    return org

end function

-----------------------------------------------------------------------------
global function getClientRect( integer self )

    -- Win32 API
    -- return client's relative rectangle
    return { 0, 0 } & ( getClientSize( self ) )
end function


-----------------------------------------------------------------------------
global function getAbsClientRect( integer self )

    -- return client's screen rectangle
    sequence org
    
    -- get origin
    org = getClientOrg( self )

    -- offset origin to get screen position
    return org & ( getClientSize( self ) + org )
    
end function



-----------------------------------------------------------------------------
function intoWindowPoint( integer self, sequence point )
    
    -- convert the point to one relative to the client
    
    if my[self][Class] = Window then
        -- relative to client origin
        point = point - getClientOrg( self )
    else
        -- relative to parent's client origin and control offset
        point = point - getClientOrg( my[self][Parent] )
        point = point - my[self][Size][X1..Y1]
    end if
    
    return point
    
end function


-----------------------------------------------------------------------------
global function getWindowRect( integer self )

    -- Win32 API
    -- returns bounding rectangle of window in screen coordinates
    sequence size 
    
    -- size of display area
    size = my[self][Size]
    
    -- window or control?
    if my[self][Class] != Window then
    
        -- adjust to parent's origin
        size[X1..Y1] = size[X1..Y1] + getClientOrg( my[self][Parent] )
            
    end if

    
    -- add origin to extent
    size[X2..Y2] = size[X2..Y2] + size[X1..Y1]
    
    return size
    
end function

-----------------------------------------------------------------------------
global function getWindowSize( integer self )

    -- Win32Lib
    -- this is used to see if a point (relative to a window) hit a
    -- control in a window
    sequence size
    
    -- get size
    size = my[self][Size]
    
    -- add origin to make wide/deep into points
    size[X2..Y2] = size[X2..Y2] + size[X1..Y1]
    
    return size
    
    
end function

-----------------------------------------------------------------------------
global function getWindowExtent( integer self )

    -- Win32Lib
    -- get extent metrics for current window
    -- returns { deep, wide }    
    return my[self][Size][X2..Y2]
    
end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Window Clipping

sequence clipStack
    clipStack = {}


-----------------------------------------------------------------------------
procedure pushClip()            
    -- save current clipping region
    clipStack = append( clipStack, clipWin )
end procedure


-----------------------------------------------------------------------------
procedure popClip()            
    -- restore saved clipping region     
    integer depth

    -- get depth of stack    
    depth = length( clipStack )
    -- restore
    clipWin = clipStack[depth]
    -- pop stack
    clipStack = clipStack[1..depth-1]
    
end procedure


-----------------------------------------------------------------------------
procedure clipTo( sequence rect )            

    --set clip window
    clipWin = { max(rect[X1],0),        max(rect[Y1],0), 
                min(rect[X2],xPixels),  min(rect[Y2],yPixels) }
                
end procedure


-----------------------------------------------------------------------------
procedure clipToClient( integer self )            

    -- sets winX and winY to the client area,
    -- and sets clipping window
    sequence rect

    -- get the absolute coordinates
    rect = getAbsClientRect( self )
    
    -- save them
    winX1 = rect[X1]
    winY1 = rect[Y1]
    winX2 = rect[X2]
    winY2 = rect[Y2]

    -- set clipping window
    clipTo( rect )    

end procedure


-----------------------------------------------------------------------------
procedure clipToWindow( integer self )            

    -- sets clipping window to window; add pixel for shadow
    clipTo( getWindowRect( self ) + {0,0,1,1} )
    
end procedure


-----------------------------------------------------------------------------
procedure clipToDesktop()
    -- set clipping window to desktop
    clipTo( { 0, 0, xPixels, yPixels } )
end procedure



-----------------------------------------------------------------------------
global function getFontSize( integer self )

    -- Win32Lib
    -- get metrics for current font

    -- return {width, height}
    -- this is hard coded for the courier 10 point
    return { 10, 16 }

end function

-----------------------------------------------------------------------------
procedure createMenu( integer self, integer owner )
    -- Win32Lib
    -- create a popup, assigned to a control
    -- attach to owner
    my[owner][MenuList] = append( my[owner][MenuList], self )
end procedure

-----------------------------------------------------------------------------
procedure createMenuItem( integer self, integer owner )
    -- Win32Lib
    -- create a submenu, assigned to a control
    my[owner][MenuList] = append( my[owner][MenuList], self )
end procedure


-----------------------------------------------------------------------------
procedure doAction( sequence actions, integer index )
    if actions[index] > -1 then
        call_proc( actions[index], {index} )
    end if
end procedure




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- drawing the desktop

-----------------------------------------------------------------------------
procedure showDesktop()

    -- redraw the desktop

    integer tileTall, self

    mouse_pointer( 0 )
    
    -- fill with solid color
    polygon( sysColor[COLOR_BACKGROUND], 
                1,  { { 0      , 0 },
                      { xPixels, 0 },
                      { xPixels, yPixels },
                      { 0      , yPixels } } )

    -- draw each window
    for this = length( windowStack ) to 1 by -1 do
        -- get window
        self = windowStack[this]
        -- visible?
        if isVisible( self ) then
            -- draw it
            showControl( self  )
        end if
    end for               
    
    mouse_pointer( 1 )
    
end procedure
  
  
  
-----------------------------------------------------------------------------
global procedure setText( integer self, sequence s )
    -- Win32
    -- set the text in control
    my[self][Text] = s
end procedure


-----------------------------------------------------------------------------
global function getText( integer self )
    -- Win32Lib
    -- return text in control
    return my[self][Text]
end function


-----------------------------------------------------------------------------
global procedure setFlag( integer self, integer bit, integer flag )
    -- Win32
    -- set the attribute flag
    
    -- turn off the flag
    if flag then           
        -- turn it on
        my[self][State] = or_bits( my[self][State], bit )
    else
        -- turn it off
        my[self][State] = and_bits( not_bits( bit ), my[self][State] )
    end if

    -- need to redraw? window refresh is handled in setVisible
    if my[self][Class] != Window then
        if  bit = IsVisible
        and not flag then
            -- control turned invisible; update parent to erase
            showControl( my[self][Parent] )
        else
            -- redraw control
            showControl( self )
        end if    
    end if

end procedure


-----------------------------------------------------------------------------
global procedure setVisible( integer self, integer flag )
    -- Win32   
    
    integer top

    -- alter the flag
    setFlag( self, IsVisible, flag )

    
    -- if window, place in proper location
    if my[self][Class] = Window then  
        
        -- get id of top item, if not self
        top = 0
        if length( windowStack ) then
            -- not self?
            if windowStack[1] != self then
                top = windowStack[1]
            end if
        end if
                    
        -- remove from the stack
        windowStack = remove( self, windowStack )

        -- setting focus?        
        if flag then       

            -- place in front
            windowStack = prepend( windowStack, self )
        
            -- need to lose focus on top item
            if top then

                -- redraw prior top window without focus
                showControl( top )
                       
                -- lost focus behavior?
                if onLostFocus[top] != -1 then
                    call_proc( onLostFocus[top], {} )
                end if
                
            end if
    
            -- draw new window on top, with focus
            showControl( self )

            -- focus behavior?
            if onGetFocus[self] != -1 then
                call_proc( onGetFocus[self], {} )
            end if

            
        else
            
            -- lose focus behavior?
            if onLostFocus[self] != -1 then
                call_proc( onLostFocus[self], {} )
            end if

            -- update the whole desktop
            showDesktop()

            -- is there a new window on top?
            if length( windowStack ) then    
            
                -- get top window
                top = windowStack[1]
                
                -- gain focus behavior?
                if onGetFocus[top] != -1 then
                    call_proc( onGetFocus[top], {} )
                end if
                
            end if

            -- has the main window (or a modal window) closed?                 
            if self = mainWindow then
                exitEventLoop = 1
            end if               
            
        end if

    end if
    
end procedure

-----------------------------------------------------------------------------
global procedure setEnable( integer self, integer flag )
    -- Win32
    setFlag( self, IsEnabled, flag )
end procedure

-----------------------------------------------------------------------------
procedure setPressed( integer self, integer flag )
    -- internal
    setFlag( self, IsPressed, flag )
end procedure


-----------------------------------------------------------------------------
global procedure setCheck( integer self, integer flag )
    -- Win32     
    setFlag( self, IsChecked, flag )
end procedure


-----------------------------------------------------------------------------
-- NEW! 0.3 renamed from unloadWindow to closeWindow
global procedure closeWindow( integer self )

    -- Win32Lib  

    -- hide the window
    setVisible( self, 0 )    
    
end procedure


-----------------------------------------------------------------------------
global function create( integer kind,
                 sequence title,
                 integer parent,
                 integer x1, integer y1,
                 integer x2, integer y2,
                 atom flags )

    -- Win32Lib
    -- create a control with given attributes

    integer self                 
    sequence attribute
    
    -- hot keys are not currently supported: remove the '&'
    title = remove( '&', title )

    -- create actions
    -- these go first in case any actions are triggered by the code below
    onOpen  = append( onOpen,  -1 )
    onPaint = append( onPaint, -1 )
    onClick = append( onClick, -1 )
    onMouse = append( onMouse, -1 )
    onRightClick = append( onRightClick, -1 )    
    onScroll = append( onScroll, -1 )
-- NEW! 0.14    
    onKeyDown = append( onKeyDown, -1 )
    onKeyPress = append( onKeyPress, -1 )
    onGetFocus = append( onGetFocus, -1 )
    onLostFocus = append( onLostFocus, -1 )    
    onTimer = append( onTimer, -1 )    

    -- default?
    if x1 = Default then x1 = floor( (xPixels - x2)/2 )  end if
    if y1 = Default then y1 = floor( (yPixels - y2)/2 )  end if
    if x2 = Default then x2 = 1  end if
    if y2 = Default then y2 = 1  end if

    if kind = List then
        -- size the list. assumes letters are 16 bits high
        y2 = y2 - y1                -- subtract top for height
        y2 = floor( y2/16 ) * 16    -- 16 bits high
        y2 = y2 + y1                -- add offset back in
        y2 = y2 + 4                 -- 2 pixels on top and bottom
    end if
    
    -- build the values
    attribute = repeat( 0, MySize )
    attribute[Class]        = kind
    attribute[Text]         = title
    attribute[Parent]       = parent
    attribute[Size]         = {x1,y1,x2,y2}
    attribute[Flags]        = flags
    attribute[MenuList]     = {}
    attribute[PenPos]       = {0,0}
    attribute[PenColor]     = sysColor[COLOR_WINDOWTEXT]
    attribute[BackColor]    = sysColor[COLOR_APPWORKSPACE]    
    attribute[FontAttrib]   = 0
    attribute[Font]         = SysFont
    attribute[Range]        = {1,100,1,10}
    attribute[State]        = IsVisible+IsEnabled
    attribute[MousePtr]     = IDC_ARROW
    attribute[Focus]        = 0
    attribute[TabStops]     = {}
    attribute[PointerAt]    = 0            
    
    -- add to my
    my = append( my, attribute )           
    self = length( my )

    -- is this a window?
    if attribute[Class] = Window then
        -- not visible
        setVisible( self, 0 )
        
    elsif attribute[Class] = Menu then
        -- create a menu    
        createMenu( self, parent )
        
    elsif attribute[Class] = MenuItem then
        -- create a menu item
        createMenuItem( self, parent )
        
    else                                   
        -- if not a static control
        if not find( kind, StaticControls ) then
            -- add to tabstops
            my[parent][TabStops] = append( my[parent][TabStops], self )            
            
            -- set as focus?
            if my[parent][Focus] = 0 then
                my[parent][Focus] = self
            end if
            
        end if
        
        
        -- display the control
        if classShow[kind] != -1 then
            call_proc( classShow[kind], {self} )                
        end if
        
        
    end if        

    -- return position
    return self
    
end function


-----------------------------------------------------------------------------
procedure putText( integer x, integer y, sequence text, 
                    integer fore, integer back, integer attrib )

    -- display
    clipImage( {x,y}, build_text_image( text, fore, back, attrib ) )

end procedure


-----------------------------------------------------------------------------
function getTopWindow( integer self )
    -- Win32
    -- returns parent window, or top window if 0
    -- this will blow up if the window stack is empty.
    -- on the other hand, returning a 0 will cause the caller to choke.
    if self = 0 then
        return windowStack[1]
    else
        return my[self][Parent]
    end if
    
end function



-----------------------------------------------------------------------------
function getFocus()
    -- Win32Lib
    -- return control with focus
    return my[getTopWindow(0)][Focus]    
end function


-----------------------------------------------------------------------------
function hasFocus( integer self )
    -- Win32Lib
    -- true if control has focus
    integer topWin, parent
    
    if my[self][Class] = Window then
        return (self = getTopWindow(0))
    else            
        -- does parent window have focus?
        parent = my[self][Parent]
        if parent = getTopWindow(0) then 
            -- see if control had focus
            return (self = getFocus())
        else
            -- no focus if not in top window
            return 0
        end if
    end if
    
end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Image Class

procedure showToggle( integer self, sequence images )

    -- draw a checkbox or radio button

    integer x1, y1, x2, y2, flags, parent
    integer index, fColor, bColor, active
    sequence image, at

    -- set anchor to parent
    parent = my[self][Parent]
    
    -- set clipping
    pushClip()
    clipToClient( parent )
    
    -- get position
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]
    
    -- get state flags
    flags = my[self][State]
    
    -- select image    
    if isPressed( self ) then
        if isChecked( self ) then
            index = PressedChecked
        else
            index = PressedUnchecked
        end if    
    elsif isEnabled( self ) then
        if isChecked( self ) then
            index = ActiveChecked
        else
            index = ActiveUnchecked
        end if
    else
        if isChecked( self ) then
            index = InactiveChecked
        else
            index = InactiveUnchecked
        end if    
    end if

    -- build image    
    parent = my[self][Parent]
    bColor = my[parent][BackColor]
    image = buildImage( images[index], BLACK, bColor )
    
    -- erase prior focus rectangle?
    if not hasFocus( self ) then           
        -- focus rectangle
        clipRect( bColor, 0, x1, y1, x2, y2 )
    end if
    
    -- draw it               
    at = justifyAt( x1, y1, x2, y2, 0, image )
    clipImage( at, image )
    
    -- the text color
    fColor = iif( isEnabled( self ), BLACK, sysColor[COLOR_WINDOWTEXT] )
    
    -- the text  
    image = build_text_image( my[self][Text], fColor, bColor, BOLD )
    at = justifyAt( x1+16, y1, x2, y2, 0, image )    
    clipImage( at, image )
    
    -- focus rectangle?
    if hasFocus( self ) then           
        -- focus rectangle
        clipFocusRect( bColor, x1, y1, x2, y2 )
    end if
    
    -- restore                
    popClip()
        
end procedure



-----------------------------------------------------------------------------
function pointInRect( sequence point, sequence rect )
    -- returns true if point is in rectangle
    
    return ( point[X1] >= rect[X1] and
             point[Y1] >= rect[Y1] and
             point[X1] <= rect[X2] and
             point[Y1] <= rect[Y2] )
             
end function



-----------------------------------------------------------------------------
function clickGeneric( integer self )

    -- action to take when generic control is clicked
    -- returns { flag, mouseX, mouseY }
    
    integer inside, flag
    sequence mouse, rect

    -- get control's rectangle
    rect = getWindowRect( self )

    while 1 do
        -- poll mouse
        mouse = poll_mouse()

        -- exit if mouse is released        
        if not and_bits( mouse[MOUSE_BUTTONS], MOUSE_LEFT ) then
            exit
        end if

        -- mouse in the button?
        inside = pointInRect( mouse[MOUSE_X..MOUSE_Y], rect )

        -- need to change state?
        if inside and not isPressed( self ) then
            -- toggle on
            setPressed( self, 1 )
            
        elsif not inside and isPressed( self ) then
            -- toggle off
            setPressed( self, 0 )
            
        end if
           
    end while 

    if isPressed( self ) then
        flag = 1
        setPressed( self, 0 )
    else
        flag = 0
    end if      


    
    return { flag } &  mouse[MOUSE_X..MOUSE_Y]
    
end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Event Handler Tools


-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- LText Class

-----------------------------------------------------------------------------
function breakText( sequence text, sequence delimiter )
    
    -- break text into segments based on delimiter
    -- ex:  breakText( "hi there\n\how are you?", "\n" )
    --      --> { "hi there", "how are you?" }

    integer breakAt    
    sequence segments

    -- clear result
    segments = {}
    
    while length( text ) do
    
        -- look for break delimiter
        breakAt = match( delimiter, text )
        if breakAt then            
            -- break into segment
            segments = append( segments, text[1..breakAt-1] )
            -- get text after break
            text = text[breakAt+length(delimiter)..length(text)]
        else
            -- no delimiters left
            segments = append( segments, text )
            -- no text remaining
            text = ""
        end if

    end while
    
    return segments
    
end function

-----------------------------------------------------------------------------
function getTextRect( sequence text, integer attributes )
    -- returns rectangle needed to enclose text
    integer x, y, wide, widest
    
    -- convert to multiple lines
    text = breakText( text, {13,10} )
    
    -- get number of lines
    y = length( text ) * 16 -- HARD CODED!
    
    -- get longest line    
    widest = 0
    for i = 1 to length( text ) do
        -- get line's width
        wide = get_text_size( text[i], attributes )
        -- save if widest
        if wide > widest then
            widest = wide
        end if
    end for
    
    return { 0, 0, widest, y }
    
end function


    
-----------------------------------------------------------------------------
procedure showLText( integer self )

    integer parent, bColor, fColor, breakAt, x, y
    sequence text

    -- get parent
    parent = my[self][Parent]                  

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if
                  
    -- set anchor to parent
    pushClip()
    clipToClient( parent )
    
    -- use parent's background color
    bColor = my[parent][BackColor]
    fColor = BLACK
    
    -- get position
    x = winX1 + my[self][Size][X1]
    y = winY1 + my[self][Size][Y1]

    -- get text
    text = my[self][Text]

    -- break at cr/lf
    text = breakText( text, {13,10} )
    
    for i = 1 to length( text ) do
        -- display segment
        putText( x, y, text[i], fColor, bColor, 0 )
        -- increment position (hard coded!)
        y = y + 16        
    end for
        
    -- return clip
    popClip()

end procedure


-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ LText ]     = routine_id( "showLText" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- RText Class

-----------------------------------------------------------------------------
procedure showRText( integer self )

    integer parent, bColor, fColor, breakAt, x, y, wide
    sequence text, image

    -- get parent
    parent = my[self][Parent]                  

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if
                  
    -- set anchor to parent
    pushClip()
    clipToClient( parent )
    
    -- use parent's background color
    bColor = my[parent][BackColor]
    fColor = BLACK
    
    -- get position
    wide = my[self][Size][X2]
    x = winX1 + my[self][Size][X1]
    y = winY1 + my[self][Size][Y1]

    -- get text
    text = my[self][Text]

    -- break at cr/lf
    text = breakText( text, {13,10} )
    
    for i = 1 to length( text ) do                             
        -- create text
        image = build_text_image( text[i], fColor, bColor, 0 )
        
        -- display segment
        putText( x+wide-length(image[1]), y, text[i], fColor, bColor, 0 )
        
        -- increment position (hard coded!)
        y = y + 16        
    end for
        
    -- return clip
    popClip()

end procedure


-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ RText ]     = routine_id( "showRText" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Push Button and Default Push Button Class

-----------------------------------------------------------------------------
procedure showPushButton( integer self )

    -- draw a command button
    integer x1, y1, x2, y2, fColor, bColor, parent, focus, default
    sequence text, image, at

    -- get parent
    parent = my[self][Parent]    

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]   

    -- get background color
    bColor = my[parent][BackColor]

    -- focus?
    focus = hasFocus( self )
    
    -- default?
    default = iif( my[self][Class] = DefPushButton, 1, 0 )

    if isPressed( self ) then
    
        -- erase edge
        clipLine( bColor, x1, y1, x2, y1 )
        clipLine( bColor, x1, y1, x1, y2 )
        
        -- button face
        clipRect( BLACK, 1, x1+1, y1+1, x2, y2 )


        -- title
        text = my[self][Text]
    
        -- build the text
        fselect( SysFont )      
        image = build_text_image( text, BRIGHT_WHITE, BLACK, BOLD )    
        
        -- display it    
        at = justifyAt( x1, y1, x2, y2, CenterJustify, image )
        clipImage( at, image )

    else    
                 
        bColor = sysColor[COLOR_BTNFACE]
        clipFramedRect( bColor, x1, y1, x2-1, y2-1 )
        
        -- shadow             
        fColor = sysColor[COLOR_BTNSHADOW]
        clipLine( fColor, x2, y1+1, x2, y2 )
        clipLine( fColor, x1+1, y2, x2, y2 )

        -- extra shadow for default?        
        if default then
            clipRect( BLACK, 0, x1+1, y1+1, x2-2, y2-2 )
        end if
        
        -- focus rect?
        if focus then
            if default then
                clipFocusRect( bColor, x1+2, y1+2, x2-3, y2-3 )
            else                
                clipFocusRect( bColor, x1+1, y1+1, x2-2, y2-2 )
            end if
        end if
                       
        -- title
        text = my[self][Text]
    
        -- the text color
        fColor = iif( isEnabled( self ), BLACK, GRAY )
        
        -- build the text
        fselect( SysFont )      
        image = build_text_image( text, fColor, bColor, BOLD )    
        
        -- display it    
        at = justifyAt( x1, y1, x2, y2, CenterJustify, image )
        clipImage( at, image )
        
    end if

    -- restore
    popClip()
    
end procedure
    



-----------------------------------------------------------------------------
procedure clickPushButton( integer self )
    -- action to take when pushbutton control is clicked
    sequence result

    -- run click
    result = clickGeneric( self )
    
    -- chosen?
    if result[1] then
        -- action attached to mouse?
        if onClick[self] > -1 then  
            -- do action
            call_proc( onClick[self], {} )
            
        end if
    end if
    
end procedure



-----------------------------------------------------------------------------
procedure keyPushButton( integer self, integer key )

    -- action to take when pushbutton control gets a key

    -- space?
    if key = VK_SPACE then               
        -- show as pressed
        setPressed( self, 1 )
        
        -- show as released
        setPressed( self, 0 )

        -- action attached to mouse?
        if onClick[self] > -1 then  
        
            -- do action
            call_proc( onClick[self], {} )
            
        end if
    end if                               
    
end procedure



-----------------------------------------------------------------------------
-- assign behaviors             

    classShow [ PushButton ]    = routine_id( "showPushButton" )
    classShow[ DefPushButton ]  = routine_id( "showPushButton" )

    classClick[ PushButton ]    = routine_id( "clickPushButton" )
    classClick[ DefPushButton ] = routine_id( "clickPushButton" )
    classKey  [ PushButton ]    = routine_id( "keyPushButton" )
    classKey  [ DefPushButton ] = routine_id( "keyPushButton" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Radio Class  


sequence radioImages
    radioImages = repeat( {}, 6 )

--  look and feel

    radioImages[ActiveUnchecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx....xxx ",
    " xx......xx ",
    "xx........xx",
    "xx........xx",
    "xx........xx",
    "xx........xx",
    " xx......xx ",
    " xxx....xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}
    
    radioImages[ActiveChecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx....xxx ",
    " xx......xx ",
    "xx...xx...xx",
    "xx..xxxx..xx",
    "xx..xxxx..xx",
    "xx...xx...xx",
    " xx......xx ",
    " xxx....xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}

    radioImages[PressedUnchecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx////xxx ",
    " xx//////xx ",
    "xx////////xx",
    "xx////////xx",
    "xx////////xx",
    "xx////////xx",
    " xx//////xx ",
    " xxx////xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}
    
    radioImages[PressedChecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx////xxx ",
    " xx//////xx ",
    "xx///xx///xx",
    "xx//xxxx//xx",
    "xx//xxxx//xx",
    "xx///xx///xx",
    " xx//////xx ",
    " xxx////xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}


    radioImages[InactiveUnchecked] = {
    "    XXXX    ",
    "  XXXXXXXX  ",
    " XXX////XXX ",
    " XX//////XX ",
    "XX////////XX",
    "XX////////XX",
    "XX////////XX",
    "XX////////XX",
    " XX//////XX ",
    " XXX////XXX ",
    "  XXXXXXXX  ",
    "    XXXX    "}
    
    radioImages[InactiveChecked] = {
    "    XXXX    ",
    "  XXXXXXXX  ",
    " XXX////XXX ",
    " XX//////XX ",
    "XX///XX///XX",
    "XX//XXXX//XX",
    "XX//XXXX//XX",
    "XX///XX///XX",
    " XX//////XX ",
    " XXX////XXX ",
    "  XXXXXXXX  ",
    "    XXXX    "}
    


-----------------------------------------------------------------------------
procedure showRadio( integer self )

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    
    -- draw button
    showToggle( self, radioImages )
    
end procedure

-----------------------------------------------------------------------------
procedure clickRadio( integer self )
    -- action to take when radio control is pressed
    sequence result

    -- run click
    result = clickGeneric( self )
    
    -- chosen?
    if result[1] then
    
        -- turn on radio
        setCheck( self, 1 )
        
        -- needs to turn off others
    
        -- action attached to mouse?
        if onClick[self] > -1 then  
        
            -- do action
            call_proc( onClick[self], {} )
            
        end if
    end if
    
end procedure


-----------------------------------------------------------------------------
procedure keyRadio( integer self, integer key )

    -- action to take when radio button control gets a key

    -- space?
    if key = VK_SPACE then               
        -- toggle radio
        setCheck( self, not( isChecked( self ) ) )
        
        -- needs to turn off others
    
        -- action attached to mouse?
        if onClick[self] > -1 then  
        
            -- do action
            call_proc( onClick[self], {} )
            
        end if
    end if                               
    
end procedure


-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ Radio ]     = routine_id( "showRadio" )
    classClick[ Radio ]    = routine_id( "clickRadio" )
    classKey[ Radio ]       = routine_id( "keyRadio" )







-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Check Box Class  

sequence checkImages
    checkImages = repeat( {}, 6 )

    checkImages[ActiveUnchecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[ActiveChecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx.........xx",
    "xx.......x.xx",
    "xx......xx.xx",
    "xx.x...xxx.xx",
    "xx.xx.xxx..xx",
    "xx.xxxxx...xx",
    "xx..xxx....xx",
    "xx...x.....xx",
    "xx.........xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[PressedUnchecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[PressedChecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx/////////xx",
    "xx///////x/xx",
    "xx//////xx/xx",
    "xx/x///xxx/xx",
    "xx/xx/xxx//xx",
    "xx/xxxxx///xx",
    "xx//xxx////xx",
    "xx///x/////xx",
    "xx/////////xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[InactiveUnchecked] = {
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX"}

    checkImages[InactiveChecked] = {
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX",
    "XX/////////XX",
    "XX///////X/XX",
    "XX//////XX/XX",
    "XX/X///XXX/XX",
    "XX/XX/XXX//XX",
    "XX/XXXXX///XX",
    "XX//XXX////XX",
    "XX///X/////XX",
    "XX/////////XX",
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX"}

-----------------------------------------------------------------------------
procedure showCheck( integer self )

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if


    showToggle( self, checkImages )
    
    
        
end procedure

-----------------------------------------------------------------------------
procedure clickCheck( integer self )
    sequence result

    -- run click
    result = clickGeneric( self )
    
    -- chosen?
    if result[1] then
    
        -- toggle
        setCheck( self, not isChecked( self ) )
    
        -- action attached to mouse?
        if onClick[self] > -1 then  
        
            -- do action
            call_proc( onClick[self], {} )
            
        end if
    end if
end procedure


-----------------------------------------------------------------------------
procedure keyCheck( integer self, integer key )

    -- action to take when check box control gets a key

    -- space?
    if key = VK_SPACE then               
    
        -- toggle check
        setCheck( self, not( isChecked( self ) ) )
    
        -- action attached to mouse?
        if onClick[self] > -1 then  
        
            -- do action
            call_proc( onClick[self], {} )
            
        end if
    end if                               
    
end procedure


-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ CheckBox ]       = routine_id( "showCheck" )
    classClick[ CheckBox ]      = routine_id( "clickCheck" )
    classKey[ CheckBox ]        = routine_id( "keyCheck" )







-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- HScroll/VScroll Classes


-----------------------------------------------------------------------------
function calcThumbRect( integer self )

    -- returns rectangle containing the thumb
    integer size, wide, at, range
    sequence rect

    rect = getWindowRect(self)
    
    -- size of scroll area
    if my[self][Class] = HScroll then
        size = rect[X2] - ( rect[X1]+6 )
    else
        size = rect[Y2] - ( rect[Y1]+6 )
    end if

    -- total range of scroll bar
    range = (my[self][Range][ScrollMax]-my[self][Range][ScrollMin])+1

    -- width is relative to large (page) scroll value
    wide = floor( size * ( my[self][Range][ScrollLarge]/range ))
    
    -- too small?
    if wide < 8 then
        wide = 8
    end if

    

    -- get relative position
    at = floor((my[self][PointerAt]/range) * (size-wide) ) + 3

    -- what type of scroll bar?
    if my[self][Class] = HScroll then
        -- horizontal
        return { rect[X1]+at, rect[Y1]+3, rect[X1]+at+wide, rect[Y2]-3 }
    else             
        -- vertical
        return { rect[X1]+3, rect[Y1]+at, rect[X2]-3, rect[Y1]+at+wide }
    end if
    
end function



-----------------------------------------------------------------------------
procedure showScroll( integer self )

    -- Redraws the scrollbar, less buttons.
    -- This is the "real time" update code, so the
    -- thumb image is pre-calculated
    
    integer parent, focus, color, x1, y1, x2, y2
    sequence thumb

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )
    
    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    ---------------------
    -- draw the scrollbar
    ---------------------

    color = sysColor[COLOR_SCROLLBAR]

    -- draw the channel
    clipFramedRect( color, x1, y1, x2, y2 )
    
    if hasFocus( self ) then
        -- draw a focus rectangle
        clipFocusRect( color, x1+1, y1+1, x2-1, y2-1 )
    end if
        
    -----------------
    -- draw the thumb
    -----------------
    
    thumb = calcThumbRect( self )              
    color = sysColor[COLOR_BTNFACE]
    clipShadowedRect( color, thumb[X1], thumb[Y1], thumb[X2], thumb[Y2] )

    -- restore
    popClip()
    
end procedure


-----------------------------------------------------------------------------
global procedure setScrollPos( integer self, integer pos )

    --$ Win32Lib Emulation
    -- updates the scroll bar to a new value, and redraws it
    -- only the part of the scroll bar that changed is updated
    
    integer parent, focus, color, x1, y1, x2, y2
    sequence thumb

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if
    
    -- keep in range
    if pos < my[self][Range][ScrollMin] 
        then pos = my[self][Range][ScrollMin]
    elsif pos > my[self][Range][ScrollMax] then
        pos = my[self][Range][ScrollMax]
    end if
    
    -- same as old value?
    if pos = my[self][PointerAt] then
        -- don't update
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )
    
    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]


    -- erase the old thumb
    color = sysColor[COLOR_SCROLLBAR]
    thumb = calcThumbRect( self )
    clipRect( color, 1, thumb[X1], thumb[Y1], thumb[X2]+1, thumb[Y2]+1 )
    
    -- update with new value     
    my[self][PointerAt] = pos
    
    -- draw new thumb
    thumb = calcThumbRect( self )
    color = sysColor[COLOR_BTNFACE]
    clipShadowedRect( color, thumb[X1], thumb[Y1], thumb[X2], thumb[Y2] )
        
    -- restore
    popClip()
    
    -- action?
    if onScroll[self] != -1 then
        call_proc( onScroll[self], {my[self][PointerAt]} )
    end if
    
    
end procedure


-----------------------------------------------------------------------------
procedure clickScroll( integer self )

    -- handle clicking in scroll bar    
    integer wide, range, pos, sMin, sMax, relPos, startPos, thumbWide
    integer mouseAxis, axis1, axis2
    atom ratio
    sequence mouse, thumb, rect, priorMouse

    -- which scroll bar are we talking about?
    if my[self][Class] = HScroll then
        -- use horizontal coordinates
        axis1 = X1
        axis2 = X2
        mouseAxis = MOUSE_X
    else
        -- use vertical coordinates
        axis1 = Y1
        axis2 = Y2
        mouseAxis = MOUSE_Y
    end if

    -- get position of mouse
    mouse = poll_mouse()

    -- get position of thumb
    thumb = calcThumbRect( self )
    
    -- clicked before thumb?
    if mouse[mouseAxis] < thumb[axis1] then
    
        -- scroll back by a page 
        pos = my[self][PointerAt] - my[self][Range][ScrollLarge]

        -- display
        setScrollPos( self, pos )
        
        
    elsif mouse[mouseAxis] > thumb[axis2] then
        -- scroll ahead by a page
    
        pos = my[self][PointerAt] + my[self][Range][ScrollLarge]

        -- display
        setScrollPos( self, pos )
        
    
    else
        -- allow the user to drag the thumb around

        -- get absolute (screen) position of the controls        
        rect = getWindowRect( self )

        -- get thumb size
        thumb = calcThumbRect( self )
        thumbWide = thumb[axis2] - thumb[axis1] + 1
        
        -- range of scroll area
        wide = rect[axis2] - ( rect[axis1]+6+thumbWide )
    
        -- total range of scroll bar
        range = (my[self][Range][ScrollMax]-my[self][Range][ScrollMin])+1
        
        -- ratio of movement
        ratio = range/wide
        
        -- range
        sMin = my[self][Range][ScrollMin]
        sMax = my[self][Range][ScrollMax]
        
        -- prior position
        priorMouse = {-1,-1,-1}

        ---------------------------------------
        -- calculate relative position of mouse
        ---------------------------------------

        -- calculate start of thumb, relative to cursor
        relPos = mouse[mouseAxis]

        -- start thumb position
        startPos = my[self][PointerAt]
        
        -- loop until mouse release
        while and_bits( mouse[MOUSE_BUTTONS], MOUSE_LEFT ) do

            -- moved?
            if not compare( mouse[MOUSE_X..MOUSE_Y], 
                            priorMouse[MOUSE_X..MOUSE_Y] ) = 0 then

                -- get relative change
                pos = floor( ( (mouse[mouseAxis]-relPos)* ratio ) + startPos)
    
                -- update
                setScrollPos( self, pos )
                
                -- save mouse position
                priorMouse = mouse

            end if

            -- get the current mouse state
            mouse = poll_mouse()
            
        end while
        
    end if
    
end procedure


-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ HScroll ]        = routine_id( "showScroll" )
    classClick[ HScroll ]       = routine_id( "clickScroll" )

    classShow[ VScroll ]        = routine_id( "showScroll" )
    classClick[ VScroll ]       = routine_id( "clickScroll" )


                            

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- List Class
        
-- dummy support routines
-----------------------------------------------------------------------------
global function getIndex( integer self )
    -- dummy routine
    return 1
end function

-----------------------------------------------------------------------------
global function getCount( integer self )
    -- dummy routine
    return 1
end function

-----------------------------------------------------------------------------
global function getItem( integer self, integer index )
    -- dummy routine
    return ""
end function


-----------------------------------------------------------------------------
procedure showListBox( integer self )

    integer x1, y1, x2, y2, fColor, bColor
    sequence image, size

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )
    
    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    -- draw a filled rectangle
    bColor = sysColor[COLOR_APPWORKSPACE]
    clipFramedRect( bColor, x1, y1, x2, y2 )
    
    -- position
    x1 = x1+4
    y1 = y1 + 1

    -- populate list
    fColor = sysColor[COLOR_WINDOWTEXT]
    for i = 1 to length( my[self][Text] ) do

        -- build text image
        fselect( SysFont )
        image = build_text_image(my[self][Text][i], fColor, bColor, 0 )

        -- display
        clipImage( {x1, y1}, image )
        
        -- move down 1 line
        y1 = y1 + length( image )
        if y1 > y2 then
            exit
        end if
           
    end for

    -- should add scroll bar here
    
    
    -- restore
    popClip()
    
    
end procedure

-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ List ]       = routine_id( "showList" )






-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- EditText Class

-----------------------------------------------------------------------------
procedure showEditText( integer self )

    integer x1, y1, x2, y2, color
    sequence text, image, at

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )
    
    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]
    
    -- background color
    clipFramedRect( BRIGHT_WHITE, x1, y1, x2, y2 )
    
    -- focus?
    if hasFocus( self ) then
        -- draw focus rectangle
        clipFocusRect( BRIGHT_WHITE, x1+1, y1+1, x2-1, y2-1 )
    end if

    -- get text
    text = my[self][Text]
    if hasFocus( self ) then
        text = text & "|"
    end if

    -- generate text image
    image = build_text_image( text, BLACK, BRIGHT_WHITE, 0 )

    -- justify and draw
    at = justifyAt( x1+8, y1, x2, y2, LeftJustify, image )
    clipImage( at, image )
    
    -- restore
    popClip()
    
end procedure    


-----------------------------------------------------------------------------
procedure keyEditText( integer self, integer key )

    sequence text

    -- get the data
    text = my[self][Text]

    -- process the keystroke    
    if key = VK_TAB then
        -- this should not be added to the field
        
    elsif key = VK_BACK then
        -- any room to backspace?
        if length( text ) then
            text = text[1..length(text)-1]            
        end if
    else
        -- is there room?
        if get_text_size( text, 0 ) + 16 < my[self][Size][X2] then
            -- only allow char < 256 (or font routine will blow up)
            if key > 0 and key < 256 then
                -- add to text
                text = text & key                
            end if
        end if
    end if               
    
    -- update
    my[self][Text] = text

    -- draw the text
    showEditText( self )

end procedure


-----------------------------------------------------------------------------
-- assign behaviors             
    classShow[ EditText ]   = routine_id( "showEditText" )
    classKey[ EditText ]    = routine_id( "keyEditText" )



-----------------------------------------------------------------------------
procedure repaintWin( integer self, integer x1, integer y1,
                                    integer x2, integer y2 )
                                    
    -- redraw the workspace within the clipped area
    integer color

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- set anchor to client area
    pushClip()
    clipToClient( self )

    -- alter the clipping window to requested clip window
    clipWin = { max(winX1+x1,winX1), 
                max(winY1+y1,winY1), 
                min(winX1+x2,winX2), 
                min(winY1+y2,winY2) }

    -- hide the mouse
    mouse_pointer( 0 )
               
    -- clear client area
    color = my[self][BackColor]
    clipRect( color, 1, clipWin[X1], clipWin[Y1], clipWin[X2], clipWin[Y2] )

    -- on paint
    if onPaint[self] > -1 then
        
        -- run routine
        call_proc( onPaint[self], {x1,y1,x2,y2} )
        
    end if
                            
    -- draw controls
    for i = 1 to length( my ) do
        -- parent of non-window control?
        if  my[i][Parent] = self
        and my[i][Class] != Window then
            -- draw it
            showControl( i )            
        end if
    end for

    -- show the mouse
    mouse_pointer( 1 )
    
    -- restore clipping
    popClip()    
        
end procedure


-----------------------------------------------------------------------------
global procedure repaintWindow( integer self )

    -- Win32Lib
    -- redraw the client area
    sequence point

    -- get extent
    point = getClientSize( self )
    
    -- repaint the entire window   
    repaintWin( self, 0, 0, point[X1], point[Y1] )
        
end procedure


-----------------------------------------------------------------------------
global procedure repaintRect( integer self, integer x1, integer y1,
                                integer x2, integer y2 )

    -- Win32Lib
    -- redraw the workspace within range

    -- repaint only requested portion of window
    repaintWin( self, x1, y1, x2, y2 )
        
end procedure



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Bitmap class

sequence resource
resource = {}


-----------------------------------------------------------------------------
procedure showBitmap( integer self )

    -- draw a bitmap control
    integer parent
    sequence rect

    -- get parent
    parent = my[self][Parent]
    
    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- no bitmap?
    if my[self][Bmp] = 0 then
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- get screen position
    rect = getWindowRect( self )

    -- display the bitmap resource
    clipImage( rect[X1..Y1], resource[my[self][Bmp]] )

    -- restore
    popClip()
    
end procedure



------------------------------------------------------------------------------
function convertToPalette( sequence sourcePal )

    -- convert the color list to palette indexes
    
    atom err, diff, best
    sequence destPal, colorIndex
    
    -- get palette
    destPal = get_all_palette()
    
    -- adjust source
    sourcePal = floor( sourcePal/4 )
    
    -- build holder for results
    colorIndex = repeat( 0, length( sourcePal ) )

    -- look at each entry    
    for i = 1 to length( sourcePal ) do
        -- set the last best match as impossibly bad
        best = #FFFFFF
        -- look through the palette
        for entry = 1 to length( destPal ) do
            -- error amount for this entry
            err = 0                       
            -- for each {rgb}
            for rgb = 1 to 3 do
                -- error is difference between the two
                diff = abs( sourcePal[i][rgb] - destPal[entry][rgb] )
                err = err + diff
            end for

            -- better match than prior best match?
            if err < best then
                -- save error amount
                best = err               
                -- store the palette index
                colorIndex[i] = entry
            end if
        end for
    end for

    return colorIndex
         
end function


------------------------------------------------------------------------------
global function rgb( integer r, integer g, integer b )
    -- return index of closest match in palette
    sequence s
    s = convertToPalette( {r,g,b} )
    return s[1]
end function


-----------------------------------------------------------------------------
global function loadBitmapFromFile( sequence fName )
    -- Win32Lib
    -- load bitmap, return resource id                 
    integer index
    sequence replace
    object res

    -- attempt to read bitmap
    res = read_bitmap( fName )
    if integer( res ) then
        abortErr( "Unable to read bitmap %s", {fName}  )
    end if

    -- adjust the current palette
    replace = convertToPalette(res[1])

    
    -- replace values with references to new palette
    for i = 1 to length( res[2] ) do
        for j = 1 to length( res[2][i] ) do
            -- get index
            index = res[2][i][j]
            -- look up replacement
            index = replace[index+1]-1
            -- replace
            res[2][i][j] = index       
        end for
    end for                       
    
    -- add to resource list    
    resource = append( resource, res[2] )

    -- return index in resource list    
    return length( resource )
    
end function


-----------------------------------------------------------------------------
global procedure setBitmap( integer self, object bitmap )
    -- load bitmap
    if sequence( bitmap ) then
        -- assume file name
        my[self][Bmp] = loadBitmapFromFile( bitmap )
    else
        -- resource id
        my[self][Bmp] = bitmap        
    end if       
    
    -- update the control
    showBitmap( self )
    
end procedure


-----------------------------------------------------------------------------
global procedure drawBitmap( integer self, integer resId, 
                                    integer x, integer y )
    -- $Win32Lib
    
    -- anchor to window
    pushClip()
    clipToClient( self )

    -- draw the bitmap
    clipImage( {winX1+x+1,winY1+y+1}, resource[ resId ] )

    -- restore
    popClip()
                                    
end procedure

-----------------------------------------------------------------------------
-- NEW! 0.4 renamed from createBitmap to textToBitmap
global function textToBitmap( sequence s )
    -- create a bitmap, adding it to the resource list
    sequence res

    -- convert to colors
    for i = 1 to length( s ) do
        for j = 1 to length( s[i] ) do
            if    s[i][j] = 'x' then s[i][j] = BLACK
            else                  
                -- color of client area
                s[i][j] = sysColor[COLOR_APPWORKSPACE]
            end if
        end for
    end for
    
    -- create a resource
    resource = append( resource, s )
    
    -- return index to the resource
    return length( resource )
    
end function

-----------------------------------------------------------------------------
-- assign behaviors             

    classShow [ Bitmap ]     = routine_id( "showBitmap" )






-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Icon class

-----------------------------------------------------------------------------
procedure showIcon( integer self )

    -- draw an icon control
    -- this is a stub, since I don't know how to read icons
    sequence rect

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )

    -- get size
    rect = getWindowRect( self )

    -- place rectangle as marker
    clipFramedRect( BRIGHT_WHITE, rect[X1], rect[Y1], rect[X2], rect[Y2] )
    clipLine( BLACK, rect[X1], rect[Y1], rect[X2], rect[Y2] )
    clipLine( BLACK, rect[X2], rect[Y1], rect[X1], rect[Y2] )

    -- restore
    popClip()
    
end procedure


-----------------------------------------------------------------------------
global procedure setIcon( integer id, object icon )
    -- do-nothing stub
end procedure


-----------------------------------------------------------------------------
-- assign behaviors             

    classShow [ Icon ]     = routine_id( "showIcon" )






-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Graphics Support


-----------------------------------------------------------------------------
-- NEW! 0.4 added fill parameter
global procedure drawRectangle( integer self, integer fill, 
                                    integer x1, integer y1,
                                    integer x2, integer y2 )

    -- anchor to window
    pushClip()
    clipToClient( self )

    -- adjust to window
    x1 = winX1 + x1 - 1
    y1 = winY1 + y1 - 1
    x2 = winX1 + x2 - 1
    y2 = winY1 + y2 - 1
    
    -- draw the rectangle
    clipRect( my[self][PenColor], fill, x1, y1, x2, y2 )
    
    -- restore
    popClip()
    
end procedure                                    



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- MenuItem class



-----------------------------------------------------------------------------
-- graphics

constant
    MenuCheck = {
    "      x",
    "     xx",
    "xx  xx ",
    "xx xx  ",
    "xxxx   ",
    "xxx    ",
    "xx     "}



-----------------------------------------------------------------------------
procedure sizeDropDown( integer self, integer x1, integer y1 )

    -- assign position of the menu drop down items based on
    -- the position of the menu bar
    
    integer     item, fontHigh, menuWide, menuHigh, width, half, x2, y2
    sequence    menuList

    -- get the menu items list
    menuList = my[self][MenuList]
    
    -- select the system font
    fselect( SysFont )
    
    -- get the height of the font
    fontHigh = 16 -- HARD CODED!
    half = floor( fontHigh/2 )

    -- add border to menu
    y1 = y1 + 2
    
    -- find the size of the menu
    menuWide = 0    
    menuHigh = 0
    for i = 1 to length( menuList ) do        
        
        -- get an item
        item = menuList[i]
        -- get the width
        width = get_text_size( my[item][Text], BOLD )
        -- save if largest encountered so far
        if width > menuWide then
            menuWide = width
        end if

        -- seperators only take up half space        
        if compare( my[item][Text], "-" ) = 0 then
            y2 = y1 + half + 2
        else
            y2 = y1 + fontHigh + 2
        end if
        
        -- store values
        my[item][Size] = {x1, y1, 0, y2 }
        
        -- move ahead
        y1 = y2
        
    end for
    
    -- add in room for check box and edges
    x2 = x1 + menuWide + 20 + 4

    -- replace x2 value in list
    for i = 1 to length( menuList ) do
        -- get id
        item = menuList[i]
        -- update position
        my[item][Size][X2] = x2
    end for

end procedure


-----------------------------------------------------------------------------
procedure showMenuItem( integer self, integer pick, integer pressed )
                                
    -- draw a menu item in a dropdown list
    -- the routine sizeDropDown must be called first
                                
    integer     item, x1, y1, x2, y2, fColor, bColor, dim, half, sep
    sequence    text, image, menuList, point

    -- skip if index is zero
    if pick = 0 then
        return
    end if

    -- get the menu items list
    menuList = my[self][MenuList]

    -- get the item
    item = my[self][MenuList][pick]

    -- get attributes
    text = my[item][Text]
    x1   = my[item][Size][X1]
    y1   = my[item][Size][Y1]
    x2   = my[item][Size][X2]
    y2   = my[item][Size][Y2]

    -- is it dimmed?
    dim = not isEnabled( item )

    -- is it a seperator?
    sep = not compare( text, "-" )

    -- can't select if dim of divider
    if dim or sep then
        pressed = 0
    end if

    -- select color                      
    if dim then
        bColor = sysColor[COLOR_MENU]
        fColor = sysColor[COLOR_GRAYTEXT]
    elsif pressed then
        bColor = sysColor[COLOR_HIGHLIGHT]
        fColor = sysColor[COLOR_HIGHLIGHTTEXT]
    else
        bColor = sysColor[COLOR_MENU]
        fColor = sysColor[COLOR_MENUTEXT]
    end if

    -- handle seperator first
    if sep then

        -- mark goes at halfway point
        half = floor( (y2 - y1)/2 )
                    
        -- highlight below
        clipLine( BLACK, x1+1,y1+half,x2-1,y1+half )      

        -- exit. simplifies logic.
        return

    end if

    -- repaint the background
    clipRect( bColor, 1, x1+1, y1, x2-1, y2 )

    -- is it checked?
    if isChecked( item ) then
    
        -- build a check
        image = buildImage( MenuCheck, fColor, bColor )

        -- center against the text            
        point = justifyAt( x1+4, y1, x2, y2, 0, image )
        
        -- display it
        clipImage( point, image )
        
    end if
        
    -- build the text        
    image = build_text_image( text, fColor, bColor, BOLD )
      
    -- center against the text            
    point = justifyAt( x1+20, y1, x2, y2, 0, image )
      
    -- display the text        
    clipImage( point, image )
    
end procedure


-----------------------------------------------------------------------------
function showDropDown( integer self, integer x1, integer y1 )

    -- draw a drop-down menu, return area covered by menu
    
    integer x2, y2, last
    sequence menuList, covered

    -- calculate positions of items in dropdown
    sizeDropDown( self, x1, y1 )

    -- get the items in the dropdown
    menuList = my[self][MenuList]

    -- get the last item in the menu list
    last = menuList[length( menuList )]
    
    -- calculate the dropdown size based on the item's position
    x2 = my[last][Size][X2]
    y2 = my[last][Size][Y2]+2
                              
    -- save the covered area; account for shadow
    covered = saveImage( x1, y1, x2+1, y2+1 )
                              
    -- backdrop of the dropdown
    clipShadowedRect( sysColor[COLOR_MENU], x1, y1, x2, y2 )
    
    -- populate dropdown
    for i = 1 to length( my[self][MenuList] ) do
        showMenuItem( self, i, 0 )
    end for

    return covered

end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Menu class





-----------------------------------------------------------------------------
procedure showMenuBarItem( integer self, integer pick, integer pressed )

    -- draw a menu in a window
    integer item, x1, y1, x2, y2, fColor, bColor
    sequence image, point

    -- skip if zero
    if pick = 0 then
        return
    end if

    -- select the system font
    fselect( SysFont )

    
    if hasFocus( self ) then
        fColor = iif( pressed,  sysColor[COLOR_HIGHLIGHTTEXT], 
                                sysColor[COLOR_MENUTEXT] )
        bColor = iif( pressed,  sysColor[COLOR_HIGHLIGHT], 
                                sysColor[COLOR_MENU] )    
    else
        fColor = sysColor[COLOR_INACTIVECAPTIONTEXT]
        bColor = sysColor[COLOR_INACTIVECAPTION]
        
    end if
    
    -- build text       
    item = my[self][MenuList][pick]
    image = build_text_image( my[item][Text], fColor, bColor, Bold )

    -- get position
    x1 = my[item][Size][X1]
    y1 = my[item][Size][Y1]
    x2 = my[item][Size][X2]
    y2 = my[item][Size][Y2]
    
    -- draw box around it            
    clipRect( bColor, 1, x1, y1, x2, y2-1 )
    
    -- center against the text            
    point = justifyAt( x1+4, y1+1, x2, y2-1, 0, image )
        
    -- display text
    clipImage( point, image ) 

end procedure


-----------------------------------------------------------------------------
procedure sizeMenuBar( integer self )

    -- calculate the position of the menu bar text
    -- in the current window
    
    integer x1, y1, x2, y2, at, item
    sequence text, menuList

    -- get list of items
    menuList = my[self][MenuList]
    
    -- optional menu
    if length( menuList ) then

        -- get window position
        x1 = my[self][Size][X1]
        y1 = my[self][Size][Y1]        

        -- menu position
        x1 = x1
        y1 = y1+theTitleSize+1
        y2 = y1+theMenuSize-2
        
        -- start column
        x1 = x1 + 4

        -- build text                
        fselect( SysFont )

        -- examine each item
        for i = 1 to length( menuList ) do

            -- get the id of the menu
            item = menuList[i]
            
            -- get text
            text = my[item][Text]

            -- get width      
            x2 = x1 + get_text_size( text, Bold ) + 8

            -- store menu item position
            my[item][Size] = {x1,y1,x2,y2}

            -- move ahead
            x1 = x2 + 4
                                
        end for
        
    end if
    
end procedure


-----------------------------------------------------------------------------
procedure showMenuBar( integer self, integer pick )

    -- draw a menu in a window
    
    integer x1, y1, x2, y2, at, color, item
    sequence menuList, image, text

    -- size the items in the menu bar
    sizeMenuBar( self )

    -- get the menu items list
    menuList = my[self][MenuList]
    
    -- menu?
    if length( menuList ) then
    
        -- get menu position
        x1 = my[self][Size][X1]
        y1 = my[self][Size][Y1]+theTitleSize
        x2 = my[self][Size][X2]+x1
        y2 = y1 + theMenuSize - 1

        -- menu bar color
        if hasFocus( self ) then
            color = sysColor[COLOR_MENU]
        else
            color = sysColor[COLOR_INACTIVECAPTION]
        end if
        clipFramedRect( color, x1, y1, x2, y2 )    

        -- draw each menu item
        for i = 1 to length( menuList ) do
            -- draw the item            
            showMenuBarItem( self, i, pick=i )
        end for
    end if            
    
end procedure


-----------------------------------------------------------------------------
procedure runMenu( integer self, sequence mouseState )
    -- run the  menu
    -- point passed is relative to screen.
    
    integer pick, item, menuIndex, menuId, dropIndex, dropId, x, y,
            mouseDown
    sequence priorState, point, covered

    -- check the mouse: get the absolute position
    priorState = mouseState
    
    -- calculate the positions of items in the menu
    sizeMenuBar( self )

    -- get the point
    point = mouseState[MOUSE_X..MOUSE_Y]

    -- see if any menu bar items were hit
    menuIndex = 0
    for i = 1 to length( my[self][MenuList] ) do
    
        -- get an item
        item = my[self][MenuList][i]
        
        -- does the point fall in the item
        if pointInRect( point, my[item][Size] ) then
            -- set as index, leave loop
            menuIndex = i
            exit
        end if
    end for

    -- didn't click on any menu items?
    if menuIndex = 0 then
        return
    end if

    -- loop until the mouse is released
    mouseDown = 1
    while mouseDown do
        
        menuId = my[self][MenuList][menuIndex]
    
        -- highlight selected menu item
        showMenuBarItem( self, menuIndex, 1 )
    
        -- get the attributes of the dropdown
        dropId      = my[self][MenuList][menuIndex]    
        dropIndex   = 0
        x           = my[menuId][Size][X1]
        y           = my[menuId][Size][Y2]
    
        -- draw the dropdown, get covered area
        covered = showDropDown( dropId, x, y )
    
        -- loop until mouse released or another menu selected
        while 1 do
    
            -- check the rat
            mouseState = poll_mouse()
            
            -- mouse state changed?
            if compare( mouseState, priorState ) then
            
                -- save state
                priorState = mouseState     
            
                -- mouse release?
                if not and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) then
                    -- set flag
                    mouseDown = 0
                    -- leave the loop
                    exit
                end if
                
                -- get the point
                point = mouseState[MOUSE_X..MOUSE_Y]

                -------------------------------------------------------------
                -- menu bar hit testing
                -------------------------------------------------------------

                -- moved over a different menu item?            
                pick = 0
                    
                -- see if any menu bar items were hit
                for i = 1 to length( my[self][MenuList] ) do
                
                    -- get an item
                    item = my[self][MenuList][i]
                    
                    -- does the point fall in the item
                    if pointInRect( point, my[item][Size] ) then
                        -- set as index, leave loop
                        pick = i
                        exit
                    end if
                end for

                                
                -- selection changed?                
                if  pick != 0
                and pick != menuIndex then
                    -- erase the current dropdown
                    -- erase current menu selection
                    showMenuBarItem( self, menuIndex, 0 )
                    -- set the new item
                    menuIndex = pick
                    -- exit the inner loop
                    exit
                end if
                    
                    
                -------------------------------------------------------------
                -- dropdown hit testing
                -------------------------------------------------------------
                
                -- moved over a different dropdown item?
                pick = 0
                for i = 1 to length( my[menuId][MenuList] ) do
                    item = my[menuId][MenuList][i]
                    if pointInRect( point, my[item][Size] ) then
                        -- save, exit loop
                        pick = i
                        exit
                    end if
                end for
    
                -- selection changed?
                if pick != dropIndex then
                    -- erase prior dropdown item                                   
                    showMenuItem( dropId, dropIndex, 0 )
                    -- set the new item
                    dropIndex = pick
                    -- highlight it
                    showMenuItem( dropId, dropIndex, 1 )
                end if
                    
            end if
            
        end while

        -- restore the covered area
        restoreImage( x, y, covered )
    
        -- unhighlight menu
        showMenuBarItem( self, menuIndex, 0 )
        
    
    end while


    -- was a dropdown item selected?
    if dropIndex then              

        -- get the id of the selected item
        item = my[dropId][MenuList][dropIndex]
        
        -- active, and not a seperator?
        if  isEnabled( item )
        and compare( my[item][Text], "-" ) != 0 then
            -- run the menu code
   
            -- action attached to menu?
            if onClick[item] > -1 then  
                -- do associated action
                currentCtl = item
                call_proc( onClick[item], {} )
                currentCtl = 0
            end if
   
            
        end if
        
    end if
    

    
end procedure




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Window class

constant

GoAwayBox = {
    "xxxxxxxxxxx",
    "xXXXXXXXXXx",
    "xX        x",
    "xX        x",
    "xX        x",
    "xX        x",
    "xX        x" ,
    "xX        x",
    "xX        x",
    "xX        x",    
    "xxxxxxxxxxx" },
   
ClickedBox = {
    "xxxxxxxxxxx",
    "xXXXXxXXXXx",
    "xXx  x  x x",
    "xX x x x  x",
    "xX        x",
    "xxxx   xxxx",
    "xX        x",
    "xX x x x  x",
    "xXx  x  x x",
    "xX   x    x",    
    "xxxxxxxxxxx" },

GrowBox = {
    "xxxxxxxxxxxxxxxx",
    "x              x",
    "x              x",
    "x  xxxxxxx     x",
    "x  x     x     x",
    "x  x     xxxxx x",
    "x  x     x   x x",
    "x  x     x   x x",
    "x  x     x   x x",    
    "x  xxxxxxx   x x",
    "x    x       x x",    
    "x    x       x x",
    "x    x       x x",    
    "x    xxxxxxxxx x",
    "x              x",    
    "xxxxxxxxxxxxxxxx" },

UpArrow = {
    "xxxxxxxxxxxxxxxx",
    "x              x",
    "x              x",
    "x      xx      x",
    "x     x  x     x",
    "x    x    x    x",
    "x   x      x   x",
    "x  x        x  x",
    "x xxxx    xxxx x",
    "x    x    x    x",
    "x    x    x    x",
    "x    x    x    x",
    "x    xxxxxx    x",
    "x              x",
    "x              x",    
    "xxxxxxxxxxxxxxxx" },

DownArrow = {
    "xxxxxxxxxxxxxxxx",
    "x              x",
    "x              x",
    "x    xxxxxx    x",
    "x    x    x    x",
    "x    x    x    x",
    "x    x    x    x",
    "x xxxx    xxxx x",
    "x  x        x  x",
    "x   x      x   x",
    "x    x    x    x",
    "x     x  x     x",
    "x      xx      x",
    "x              x",
    "x              x",    
    "xxxxxxxxxxxxxxxx" },

LeftArrow = {
    "xxxxxxxxxxxxxxxxx",
    "x               x",
    "x       x       x",
    "x      xx       x",
    "x     x x       x",
    "x    x  xxxxx   x",
    "x   x       x   x",
    "x  x        x   x",
    "x  x        x   x",
    "x   x       x   x",
    "x    x  xxxxx   x",
    "x     x x       x",
    "x      xx       x",
    "x       x       x",
    "x               x",    
    "xxxxxxxxxxxxxxxxx" },

RightArrow = {
    "xxxxxxxxxxxxxxxxx",
    "x               x",
    "x      x        x",
    "x      xx       x",
    "x      x x      x",
    "x  xxxxx  x     x",
    "x  x       x    x",
    "x  x        x   x",
    "x  x        x   x",
    "x  x       x    x",
    "x  xxxxx  x     x",
    "x      x x      x",
    "x      xx       x",
    "x      x        x",
    "x               x",    
    "xxxxxxxxxxxxxxxxx" }


-----------------------------------------------------------------------------
function closeIconPos( sequence rect )
    -- return position of minimize icon relative to window
    
    -- center it vertically                  
    return justifyAt(  rect[X2]-(length(GoAwayBox)+6), 
                        rect[Y1], 
                        rect[X2], 
                        rect[Y1]+theTitleSize,
                        CenterJustify, GoAwayBox )
                        
end function                                    

    
-----------------------------------------------------------------------------
procedure xShowWindow( integer self )
        
    -- draw a window

    integer x1, y1, x2, y2, y, focus, fColor, bColor
    sequence title, image, rect, org, point

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- clip to the window's size
    pushClip()
    clipToWindow( self )            

    -- get values
    x1 = my[self][Size][X1]
    y1 = my[self][Size][Y1]
    x2 = x1+my[self][Size][X2]
    y2 = y1+my[self][Size][Y2]
    rect = {x1,y1,x2,y2}
    title = my[self][Text]

    -- color scheme based on focus
    focus = hasFocus( self)

    -- the window
    clipShadowedRect( sysColor[COLOR_WINDOW], x1, y1, x2, y2 )

    ----------------
    -- the title bar
    ----------------

    -- pick the color    
    if focus then
        fColor = sysColor[COLOR_CAPTIONTEXT]
        bColor = sysColor[COLOR_ACTIVECAPTION]
    else
        fColor = sysColor[COLOR_INACTIVECAPTIONTEXT]
        bColor = sysColor[COLOR_INACTIVECAPTION]
    end if

    -- the title bar
    clipFramedRect( bColor, x1, y1, x2, y1+theTitleSize )

    -- generate the title
    fselect( SysFont )      
    image = build_text_image( title, fColor, bColor, BOLD )

    -- center it vertically                  
    point = justifyAt( x1, y1, x2, y1+theTitleSize, CenterJustify, image )
                     
    -- display the title
    clipImage( point, image)   

    -- only show close icon is has focus and not modal
    if focus
    and not my[self][Modal] then
        -- the close icon
        clipImage( closeIconPos( rect ),
            buildImage( GoAwayBox, BLACK, BRIGHT_WHITE ) )
    end if

    -- the menu
    showMenuBar( self, 0 )

    -- call repaint to handle controls and client area
    repaintWindow( self )
    
    -- restore
    popClip()
        
end procedure


-----------------------------------------------------------------------------
procedure showWindow( integer self )
        
    -- draw a window

    integer x1, y1, x2, y2, y, focus, fColor, bColor
    sequence title, image, rect, org, point

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
        return
    end if

    -- clip to the window's size
    pushClip()
    clipToWindow( self )            

    -- get values
    x1 = my[self][Size][X1]
    y1 = my[self][Size][Y1]
    x2 = x1+my[self][Size][X2]
    y2 = y1+my[self][Size][Y2]
    rect = {x1,y1,x2,y2}
    title = my[self][Text]

    -- color scheme based on focus
    focus = hasFocus( self)

    -- the window
    clipShadowedRect( sysColor[COLOR_WINDOW], x1, y1, x2, y2 )

    ----------------
    -- the title bar
    ----------------

    -- pick the color    
    if focus then
        fColor = sysColor[COLOR_CAPTIONTEXT]
        bColor = sysColor[COLOR_ACTIVECAPTION]
    else
        fColor = sysColor[COLOR_INACTIVECAPTIONTEXT]
        bColor = sysColor[COLOR_INACTIVECAPTION]
    end if

    -- the title bar
    clipFramedRect( bColor, x1, y1, x2, y1+theTitleSize )

    -- generate the title
    fselect( SysFont )      
    image = build_text_image( title, fColor, bColor, BOLD )

    -- center it vertically                  
    point = justifyAt( x1, y1, x2, y1+theTitleSize, CenterJustify, image )
                     
    -- display the title
    clipImage( point, image)   

    -- only show close icon is has focus and not modal
    if focus
    and not my[self][Modal] then
        -- the close icon
        clipImage( closeIconPos( rect ),
            buildImage( GoAwayBox, BLACK, BRIGHT_WHITE ) )
    end if

    -- the menu
    showMenuBar( self, 0 )

    -- call repaint to handle controls and client area
    repaintWindow( self )
    
    -- restore
    popClip()
        
end procedure

    
-----------------------------------------------------------------------------
function clickWindowIcon( sequence rect, sequence iconUp, sequence iconDown )
    -- generic routine for handling mouse clicking window icons
    integer pressed
    sequence mouseState, priorState, point

    -- force icon draw
    pressed = 0
    
    -- poll the mouse
    mouseState = poll_mouse()
    
    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do

        -- check the rat
        mouseState = poll_mouse()
        priorState = {}

        -- changed?
        if compare( mouseState, priorState ) then
        
            -- save state
            priorState = mouseState
        
            -- redraw icon?
            point = mouseState[MOUSE_X..MOUSE_Y]

            -- in icon?            
            if pointInRect( point, rect ) then
                -- need to draw as pressed?
                if not pressed then
                
                    -- toggle flag
                    pressed = 1
                
                    -- redraw
                    clipImage( rect[X1..Y1],
                         buildImage( iconDown, BLACK, BRIGHT_WHITE ) )
                         
                end if
            else
                if pressed then
                    -- toggle flag
                    pressed = 0
                
                    -- redraw
                    clipImage( rect[X1..Y1],
                         buildImage( iconUp, BLACK, BRIGHT_WHITE ) )
                end if
            end if
        end if
    end while

    -- return position
    point = mouseState[MOUSE_X..MOUSE_Y]
    return pointInRect( point, rect )
    
end function


-----------------------------------------------------------------------------
procedure clickCloseIcon( integer self )

    -- handle the close icon being clicked
    sequence point

    -- get position of icon
    point = closeIconPos( getWindowRect( self ) )
    
    -- wait until mouse release
    if clickWindowIcon( point & (point+14), GoAwayBox, ClickedBox ) then
        -- close the window
        closeWindow( self )
    end if
    
end procedure



-----------------------------------------------------------------------------
-- window routine to use
    classShow[Window] = routine_id( "showWindow" )


-----------------------------------------------------------------------------
-- NEW! 0.3 renamed from loadWindow to openWindow
global procedure openWindow( integer self )

    -- Win32Lib            
    -- activate a window, put it on top

    -- on open
-- NEW! 0.4 renamed to onOpen
    if onOpen[self] > -1 then
        call_proc( onOpen[self], {} )
    end if

    -- make it visible
    setVisible( self, 1 )

end procedure

-----------------------------------------------------------------------------
-- NEW! 0.4 renamed from loadModal to openModal
global procedure openModal( integer self )

    -- make the window modal
    my[self][Modal] = 1
    
    -- load the window
    openWindow( self )

end procedure




-----------------------------------------------------------------------------             
global procedure moveWindow( integer self )

    -- handles windows dragging
    integer startX, startY, dx, dy, x, y, parent
    sequence xorRect, winRect, mouseState, priorState, parentClip

    -- full size of the window
    winRect = getWindowRect( self )

    -- check the mouse: get the absolute position
    mouseState = poll_mouse()
    priorState = mouseState
    startX = mouseState[MOUSE_X]
    startY = mouseState[MOUSE_Y]
    x = my[self][Size][X1]
    y = my[self][Size][Y1]

    -- no xor box drawn yet
    xorRect = {}
    
    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do

        -- check the rat
        mouseState = poll_mouse()
        
        -- changed?
        if compare( mouseState, priorState ) then
        
            -- save state
            priorState = mouseState
        
            -- change in position
            dx = mouseState[MOUSE_X] - startX
            dy = mouseState[MOUSE_Y] - startY

            -- calculate x, y
            x = my[self][Size][X1]+dx
            y = my[self][Size][Y1]+dy
            
            -- is there a saved region to restore?
            if length( xorRect ) then
                -- draw a border on parent
                clipXorRect( WHITE, xorRect )
                xorRect = {}
            end if
            
            -- new border position
            xorRect = winRect + {dx, dy, dx, dy}            
            clipXorRect( WHITE, xorRect )
                           
        end if
        
    end while
    
    -- is there a saved region to restore?
    if length( xorRect ) then
        -- draw a border on parent
        clipXorRect( WHITE, xorRect )
        xorRect = {}
    end if

    -- save new position?
    if startX != x
    or startY != y then

        -- change the position
        my[self][Size][X1..Y1] = {x,y}
        
        -- redraw desktop
        showDesktop()
        
    end if
                        
end procedure



-----------------------------------------------------------------------------             
global procedure resizeWindow( integer self )

    -- handles windows resizing
    integer parent, diffX, diffY
    sequence xorRect, winRect, mouseState, priorState, parentClip

    -- full size of the window
    winRect = getWindowRect( self )

    -- get difference in size
    diffX = winRect[X2] - winRect[X1] - my[self][Size][X2]
    diffY = winRect[Y2] - winRect[Y1] - my[self][Size][Y2]
    
    -- check the mouse: get the absolute position
    mouseState = poll_mouse()
    priorState = mouseState

    -- no xor box drawn yet
    xorRect = {}
    
    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do

        -- check the rat
        mouseState = poll_mouse()
        
        -- changed?
        if compare( mouseState, priorState ) then
        
            -- save state
            priorState = mouseState
        
            -- dragging top?
            if and_bits( theBorderEdge, North)
            and winRect[Y2] - mouseState[MOUSE_Y] > 100 then
                winRect[Y1] = mouseState[MOUSE_Y]
            end if

            -- dragging bottom?            
            if and_bits( theBorderEdge, South)
            and mouseState[MOUSE_Y] - winRect[Y1] > 100 then
                winRect[Y2] = mouseState[MOUSE_Y]
            end if
                               
            -- dragging left?
            if and_bits( theBorderEdge, West)
            and winRect[X2] - mouseState[MOUSE_X] > 100 then
                winRect[X1] = mouseState[MOUSE_X]
            end if

            -- dragging right?            
            if and_bits( theBorderEdge, East)
            and mouseState[MOUSE_X] - winRect[X1] > 100 then
                winRect[X2] = mouseState[MOUSE_X]
            end if

            -- is there a saved region to restore?
            if length( xorRect ) then
                -- draw a border on parent
                clipXorRect( WHITE, xorRect )
                xorRect = {}
            end if
            
            -- new border position
            xorRect = winRect
            clipXorRect( WHITE, xorRect )
                           
        end if
        
    end while
    
    -- is there a saved region to restore?
    if length( xorRect ) then
        -- draw a border on parent
        clipXorRect( WHITE, xorRect )
        xorRect = {}
    end if

    -- adjust difference
    winRect[X2] = winRect[X2] - winRect[X1] - diffX
    winRect[Y2] = winRect[Y2] - winRect[Y1] - diffY

    -- save the size
    my[self][Size] = winRect
        
    -- redraw desktop
    showDesktop()
        
end procedure




-----------------------------------------------------------------------------
global procedure addItem( integer self, sequence item )
    -- Win32Lib
    -- add item to list
    my[self][Text] = append( my[self][Text], item )
    -- redisplay
    showControl( self )
end procedure


-----------------------------------------------------------------------------
global procedure setFont( integer self, sequence name, integer points, 
                            integer attrib )

    -- Win32Lib
    -- look for a replacement to load
    my[self][Font] = 0
    for i = 1 to length(FontNames) by 2 do
            if compare( name, FontNames[i] ) = 0 then
                -- load the font
                my[self][Font] = fload( FontNames[i+1] )
                exit
            end if
    end for
    
    -- no match?
    if my[self][Font]= 0 then
        -- default
        my[self][Font] = SysFont
        
    end if
    
    -- save size and attributes
    my[self][FontSize]   = points
    my[self][FontAttrib] = attrib
        
end procedure    


-----------------------------------------------------------------------------
global procedure setDefaultFont( integer self )
    -- Win32Lib
    -- set font in window to default font
    
    -- save size and attributes
    my[self][Font]       = SysFont    
    my[self][FontSize]   = 10
    my[self][FontAttrib] = 0
    
end procedure





-----------------------------------------------------------------------------
global procedure setPosition( integer self, integer x, integer y )
    -- Win32Lib
    -- position the pen
    my[self][PenPos] = { x, y }
end procedure


-----------------------------------------------------------------------------
global procedure setPenColor( integer self, integer color )
    -- Win32Lib
    -- pen color
    my[self][PenColor] = color
end procedure

-----------------------------------------------------------------------------
global procedure setTextColor( integer self, integer color )
    -- Win32Lib
    -- same as pen color
    my[self][PenColor] = color
end procedure



-----------------------------------------------------------------------------
global procedure drawLine( integer self, integer x1, integer y1, 
                            integer x2, integer y2 )

    -- Win32Lib
    -- draw a line
    -- This does not reposition the pen.

    -- offset into window
    pushClip()
    clipToClient( self )
    
    -- draw a clipped line
    clipLine( my[self][PenColor], winX1+x1, winY1+y1, 
                                    winX1+x2, winY1+y2 )

    -- restore clipping
    popClip()    
    
    
end procedure


-----------------------------------------------------------------------------
global procedure setScrollRange( integer self, integer min, integer max  )
    --$ Win32Lib Emulation
    -- Sets the scroll bar minimum and maximum values.
    my[self][Range][1..2] = {min,max}
end procedure


-----------------------------------------------------------------------------
global procedure setScrollChange( integer self, integer line, integer page )
    --$ Win32Lib Emulation
    -- Sets the range for the scroll bar.
    -- This defines how much the scroll bar will move.    
    my[self][Range][3..4] = {line,page}
end procedure


-----------------------------------------------------------------------------
global procedure wPuts( integer self, sequence text )
    --$ Win32Lib Emulation
    -- Display text in requested window.
    -- This is the Win32Lib version of Euphoria's puts() command.
    sequence pos, image

    -- calculate position    
    pushClip()
    clipToClient( self )
    
    -- select font
    fselect( my[self][Font] )

    -- build an display text
    image = build_text_image( text,     -- text
                my[self][PenColor],     -- text color comes from pen color
                my[self][BackColor],    -- background color
                my[self][FontAttrib] )  -- text attribute
                
    -- display the text                
    clipImage( {winX1, winY1} + my[self][PenPos], image )

    -- restore clipping
    popClip()    
    
    
end procedure


-----------------------------------------------------------------------------
global procedure wPrintf( integer id, sequence s1, sequence s2 )
    --$ Win32Lib Emulation
    -- Display text in requested window.
    -- This is the Win32Lib version of Euphoria's printf() command.
    wPuts( id, sprintf( s1, s2 ) )
    
end procedure


-----------------------------------------------------------------------------
global procedure setTimer( integer window, integer id, integer ticks )
    --$ Win32Lib Emulation
    -- Activate a timer.
    -- { Timer in, TimerId, TimerTicks, TimerStart }
    timers = append( timers, { window, id, ticks/1000, time() } )
end procedure    

-----------------------------------------------------------------------------
global procedure killTimer( integer id, integer timerId )
    --$ Win32Lib Emulation
    -- Deactivate a timer.
    sequence timer
    for i = 1 to length( timers ) do
        -- get a timer
        timer = timers[i]

        -- match?        
        if timer[TimerWin] = id
        and timer[TimerId] = timerId then                       
            -- remove it
            timers = timers[1..i-1] & timers[i+1..length(timers)]
            -- finished
            return            
        end if
    end for
end procedure



-----------------------------------------------------------------------------
global procedure setWindowBackColor( integer self, integer rgb )
    --$ Win32Lib Emulation
    -- for now, hack it. will need to add a "find closest rgb" routine    
    my[self][BackColor]    = rgb
    -- needs to generate a "system color changed" event
end procedure

-----------------------------------------------------------------------------
global function getSysColor( integer index )    
    -- Win32Lib
    -- return system color
    return sysColor[ index ]
end function
    
    

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The Event Loop



-----------------------------------------------------------------------------
function inWindow( sequence point )
    -- return the id of the window the point isWin
    integer self
    sequence rect
    
    -- look through the stack
    for i = 1 to length( windowStack ) do 
    
        -- get a window
        self = windowStack[i]         
        
        -- point fallsWin window?
        if pointInRect( point, getWindowRect( self ) ) then
            return self            
        end if
        
        -- is the window modal?
        if my[self][Modal] then
            -- don't look any further
            return 0
        end if
        
    end for
    
    return 0
    
end function


-----------------------------------------------------------------------------
constant           
    InDesk      = 0,        -- not in a window
    InTitle     = 1,        -- in title bar, not on window control
    InMenu      = 2,        -- in the menu bar
    InBorder    = 3,        -- on the window border (resize)
    InContent   = 4,        -- in client area
    InMinimize  = 5,        -- on window grow icon
    InGrow      = 6,        -- on window grow icon
    InClose     = 7         -- on window close icon
    
    
-----------------------------------------------------------------------------
function inPart( integer self, sequence point )
    -- return the part of the window that was hit
    integer x, y
    sequence rect, pos

    -- get the points
    x = point[X1]
    y = point[Y1]
    
    -- get the window size    
    rect = getWindowRect( self )
    
    -- on the border?
    if x < rect[X1] + theBorderSize 
    or x > rect[X2] - theBorderSize 
    or y < rect[Y1] + theBorderSize 
    or y > rect[Y2] - theBorderSize then
    
        -- where in the border?
        theBorderEdge = 0
        
        if y < rect[Y1] + theBorderSize then
            theBorderEdge = theBorderEdge+North
        elsif y > rect[Y2] - theBorderSize then
            theBorderEdge = theBorderEdge+South
        end if

        if x < rect[X1] + theBorderSize then
            theBorderEdge = theBorderEdge+West
        elsif x > rect[X2] - theBorderSize then
            theBorderEdge = theBorderEdge+East
        end if
        
        return InBorder
        
    end if
    
    -- in the title bar?
    if y < rect[Y1] + theTitleSize then
        -- in a title bar control?
            
            -- the close icon?
            pos = closeIconPos( rect )
            if pointInRect( point, pos & (pos+14) ) then
                return InClose
            end if           
            
            -- just in the title            
            return InTitle

    end if

    -- in the menu bar?
    if hasMenu( self ) then
        if y < rect[Y1] + theTitleSize + theMenuSize then
            return InMenu
        end if
    end if

    -- must beWin the workspace
    return InContent
    
end function


-----------------------------------------------------------------------------
function inControl( integer self, sequence point )
    -- return the control that was hit, or zero if none

    -- go through controls
    for control = 1 to length( my ) do
        -- owned by window?
        if my[control][Parent] = self then
            -- hit?
            if pointInRect( point, getWindowRect( control ) ) then
            
                -- is it static?          
                if find( my[control][Class], StaticControls ) then
                    -- can't get events
                    return 0
                else
                    return control
                end if
            end if
            
        end if
        
    end for
    
    -- no matches
    return 0
    
end function


-----------------------------------------------------------------------------
procedure runTimers()

    integer id
    sequence timer 
    
    for i = 1 to length( timers ) do
        -- get a timer
        timer = timers[i]
        
        -- compare changeWin time
        if time() - timer[TimerStart] >= timer[TimerTicks] then
        
            -- reset timer
            timers[i][TimerStart] = time()
        
            -- action?
            id = timer[TimerWin]
            
            if onTimer[id] > -1 then
            
                -- pass it the timer id
                call_proc( onTimer[ id ], {timer[TimerId]} )
                
            end if
                        
        end if
        
    end for

end procedure        


-----------------------------------------------------------------------------
procedure setForegroundWindow( integer self )
    -- Win32                           
    setVisible( self, 1 )    
end procedure




-----------------------------------------------------------------------------
global procedure setFocus( integer self )

    -- Win32Lib
    integer prior, parent

    -- focus on a control
    if my[self][Class] = Window then
        setForegroundWindow( self )
        return
    end if

    -- skip if has focus    
    prior = getFocus()
    if prior = self then
        return
    end if
    
    -- lose focus        
    parent = my[self][Parent]
    my[parent][Focus] = self
    
    -- code?
    if prior then
        -- redraw control
        showControl( prior )
    
        if onLostFocus[prior] > -1 then
            call_proc( onLostFocus[prior], {} )
        end if  
        
    end if

    -- draw focused control
    setVisible( self, 1 )
    
end procedure



-----------------------------------------------------------------------------
procedure tabAhead( integer self, integer direction )

    -- tabs to nth control in window
    -- on error, does nothing
    integer tabStop
    sequence tabStops

    -- get tabs
    tabStops = my[self][TabStops]
    
    if length( tabStops ) = 0 then
        -- nothing to focus on! 
        return
    end if
    
    -- find current tab, plus direction
    tabStop = find( my[self][Focus], tabStops ) + direction
    
    -- check for wrapping
    if tabStop > length( tabStops ) then
        -- wrap to front
        tabStop = 1
        
    elsif tabStop < 1 then
        -- wrap to end
        tabStop = length( tabStops )
        
    end if

    -- find the tab
    setFocus( tabStops[ tabStop ] )
    
end procedure


-----------------------------------------------------------------------------
procedure eventLoop( integer mainWin )
    -- run the main event loop
    integer priorMain, key, hitWindow, hitPart, hitControl, class, action
    sequence point, inPoint, newKeys, oldKeys
    object mouse
    
    integer
        focusWindow,
        focusControl

    -- if there was a prior main window, save it
    priorMain = mainWindow
    
    -- set new main window
    mainWindow = mainWin
    
    -- open the window
    openWindow( mainWindow )

    -- reset flag
    exitEventLoop = 0

    -- mouse pointer
    mouse_pointer(1)

    -- clear the key queue
    newKeys = {}
    oldKeys = {}
    
    while not exitEventLoop do

        -- get the top window
        focusWindow = windowStack[1]
        focusControl = my[focusWindow][Focus]

-- NEW! 0.4 route events to the window if no controls in window
        -- if no control has focus, send event to the window        
        if focusControl = 0 then
            focusControl = focusWindow
        end if
    
    -----------------------------------------------------------------
    -- handle timer events
    -----------------------------------------------------------------
        
        if length( timers ) then
            runTimers()            
        end if

    -----------------------------------------------------------------
    -- handle key events
    -----------------------------------------------------------------
                                    

        -- check the keyboard
        -- newKeys = scanToAscii( get_keys() )
        newKeys = {get_key()}
        
        -- new key pressed?
        if compare( newKeys, oldKeys) != 0
        and compare( newKeys, {-1} ) != 0 then

            -- DEBUG! Eventually remove
            if compare( newKeys, {-1} ) != 0 then
                if compare( newKeys, {27} ) = 0 then
                    abortErr( "User Pressed Esc Key.", {} )
                end if
            end if

-- NEW! 0.4 removed conversion to Win32 mapping; use VK_xxx instead
            
            -- what keys were released?
            for i = 1 to length( oldKeys ) do
                if not find( oldKeys[i], newKeys ) then
                    -- generate a key release event
                end if                
            end for   

            -- what new keys were pressed?
            for i = 1 to length( newKeys ) do
                if not find( newKeys[i], oldKeys ) then
                    -- get the key
                    key = newKeys[i]
                    
                    if key = VK_TAB then        
                        -- tab ahead
                        tabAhead( focusWindow, 1 )
                        
                    elsif key = SHIFT_TAB then
                        -- tab back
                        tabAhead( focusWindow, -1 )
                        
                    else
                        -- position( 2, 1 )
                        -- ? key
                    end if  
            
                    -- pass key to control?
                    if focusControl then

                        -- get class
                        class = my[focusControl][Class]
                        
                        -- get action             
                        action = classKey[class]
                        
                        if action > -1 then
                            call_proc( action, {focusControl, key} )
                        end if             

-- NEW! 0.4 added onKeyDown processing
                        -- on key down event?
                        action = onKeyDown[focusControl]
                        if action > -1 then          
-- NEW! 0.4 passes only the key                        
                            call_proc( action, {key} )
                        end if             
                    
-- NEW! 0.4 onKeyPress only processes visible keys
                        -- visible key?
                        if key >= 32 and key <= 175 then
                        
                            -- look for key press event
                            action = onKeyPress[focusControl]
                            if action > -1 then
-- NEW! 0.4 passes only the key
                                call_proc( action, {key} )
                            end if             
                            
                        end if
                                            
                    else
                        -- pass key to window?
                        
                        -- on key event?
                        action = onKeyPress[focusWindow]
                        
                        if action > -1 then
                            call_proc( action, {key} )
                        end if             
                        
                        
                    end if                
                end if
            end for               
        end if

        -- update keys
        oldKeys = newKeys        
        
    -----------------------------------------------------------------
    -- handle mouse events
    -----------------------------------------------------------------
                
        -- poll mouse
        mouse = get_mouse()
        
        -- different?        
        if sequence( mouse ) then
        
            -- get point
            point = mouse[MOUSE_X..MOUSE_Y]
            hitPart = 0
            hitControl = 0
            
            -- where did it land?
            hitWindow = inWindow( point )

            -----------------
            -- mouse pointer
            -----------------
            
            -- hit window?
            if hitWindow then

                -- button press?
                if and_bits( mouse[MOUSE_EVENT], LEFT_DOWN+RIGHT_DOWN ) 
                then
                    -- needs focus?
                    if not hasFocus( hitWindow ) then
                        -- change the focus
                         setFocus( hitWindow )
                    end if
                    
                end if
                             
                -- change the cursor
                set_mouse_cursor( IDC_ARROW )
                hitPart = inPart( hitWindow, point )

                -- in client area?
                if hitPart = InContent then
                    -- determine what control, if any, was hit
                    hitControl = inControl( hitWindow, point )
                                                            
                elsif hitPart = InBorder then
                    -- select proper cursor
                    if theBorderEdge = North
                    or theBorderEdge = South then
                        set_mouse_cursor( IDC_SIZENS )
                        
                    elsif theBorderEdge = East
                       or theBorderEdge = West then
                        set_mouse_cursor( IDC_SIZEWE )
                    
                    elsif theBorderEdge = North+West
                       or theBorderEdge = South+East then
                        set_mouse_cursor( IDC_SIZENWSE ) 
                        
                    elsif theBorderEdge = South+West
                       or theBorderEdge = North+East then
                        set_mouse_cursor( IDC_SIZENESW ) 
                        
                    end if
                    
                end if
                
            else                             
                -- set pointer to indicate not in any window
                set_mouse_cursor( IDC_NO )
            end if


            ----------------
            -- onMouse event
            ----------------

            if  hitWindow
            and hitPart = InContent then

                -- point relative to hit window
                inPoint = intoWindowPoint( hitWindow, point )
                
                -- hit a control?
                if hitControl then
                    action = onMouse[hitControl]
                else
                    action = onMouse[hitWindow]
                end if

                -- is there a mouse action?                                    
                if action > -1 then
                    -- move
                    if and_bits( mouse[MOUSE_EVENT], MOVE ) then
                        call_proc( action, {MOUSE_MOVE} & inPoint )
                    end if

                    -- left down
                    if and_bits( mouse[MOUSE_EVENT], LEFT_DOWN ) then
                        call_proc( action, {LEFT_DOWN} & inPoint )
                    end if                         
                    
                    -- right down
                    if and_bits( mouse[MOUSE_EVENT], RIGHT_DOWN ) then
                        call_proc( action, {RIGHT_DOWN} & inPoint )
                    end if

                    -- left up
                    if and_bits( mouse[MOUSE_EVENT], LEFT_UP ) then
                        call_proc( action, {LEFT_UP} & inPoint )
                    end if
                
                    -- right up
                    if and_bits( mouse[MOUSE_EVENT], RIGHT_UP ) then
                        call_proc( action, {RIGHT_UP} & inPoint )
                    end if
                    
                end if
            end if
                                        
                                        
            ----------------------------
            -- mouse class action events
            ----------------------------
                
            -- left button down on something?
            if and_bits( mouse[MOUSE_EVENT], LEFT_DOWN )
            and hitWindow then               
            
                -- mouse drag                                
                if hitPart = InTitle then
                    moveWindow( hitWindow )
                    
                elsif hitPart = InBorder then
                    resizeWindow( hitWindow )
                        
                -- hit title bar?
                elsif hitPart = InTitle then
                    -- window control handler

                elsif hitPart = InClose then
                    -- close window control
                    clickCloseIcon( hitWindow )
                        
                -- hit the menu bar?
                elsif hitPart = InMenu then             
                    -- menu handler               
                    runMenu( hitWindow, mouse )

                -- hit the client area?                            
                elsif hitPart = InContent then

                    -- hit a control?                            
                    if hitControl then
                        
                        -- focus on control
                        setFocus( hitControl )
                    
                        -- get class
                        class = my[hitControl][Class]
                        
                        -- get class click action             
                        action = classClick[class]
                        if action > -1 then
                            call_proc( action, {hitControl} )
                        end if        
                                 
                                     
                    else                
                        
                        -- clicked in workspace
                        action = onClick[hitWindow]
                        if action > -1 then         
                            call_proc( action, {} )
                        end if             
                        
                    end if -- hitControl/inContent
                end if -- where hit
            end if -- MOUSE_LEFT and hitWindow
        
        end if
        
    end while
    
    -- hide mouse
    mouse_pointer(0)
    
    -- reset flag
    exitEventLoop = 0
                   
    -- restore the prior window
    mainWindow = priorMain
                   
end procedure




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--$ The Message Box
-- 
-- This emulates the message box that is available in the Win32 version
-- of Euphoria. It is a difficult control to emulate, because it returns
-- a value. That means that the message box has it's own event loop.
--
-- In practice, the message box uses the same event loop as the main
-- code, but calls and exits from it.



-----------------------------------------------------------------------------
-- Possible style values for message_box() style sequence
global constant 
    MB_ABORTRETRYIGNORE = #02, --  Abort, Retry, Ignore
    MB_APPLMODAL = #00,       -- User must respond before doing something else
    MB_DEFAULT_DESKTOP_ONLY = #20000,    
    MB_DEFBUTTON1 = #00,      -- First button is default button
    MB_DEFBUTTON2 = #100,      -- Second button is default button
    MB_DEFBUTTON3 = #200,      -- Third button is default button
    MB_DEFBUTTON4 = #300,   -- Fourth button is default button
    MB_HELP = #4000,            -- Windows 95: Help button generates help event
    
    MB_ICONHAND         = #10,
    MB_ICONQUESTION     = #20,
    MB_ICONEXCLAMATION  = #30,
    MB_ICONASTERISK     = #40,
    
    MB_ICONINFORMATION  = MB_ICONASTERISK,
    MB_ICONSTOP         = MB_ICONHAND,
    MB_ICONERROR        = MB_ICONHAND,
    MB_ICONWARNING      = MB_ICONEXCLAMATION,
    
    MB_OK = #00,              -- Message box contains one push button: OK
    MB_OKCANCEL = #01,        -- Message box contains OK and Cancel
    MB_RETRYCANCEL = #05,     -- Message box contains Retry and Cancel
    MB_RIGHT = #80000,        -- Windows 95: The text is right-justified
    MB_RTLREADING = #100000,   -- Windows 95: For Hebrew and Arabic systems
    MB_SERVICE_NOTIFICATION = #40000, -- Windows NT: The caller is a service 
    MB_SETFOREGROUND = #10000,   -- Message box becomes the foreground window 
    MB_SYSTEMMODAL  = #1000,    -- All applications suspended until user responds
    MB_TASKMODAL = #2000,       -- Similar to MB_APPLMODAL 
    MB_YESNO = #04,           -- Message box contains Yes and No
    MB_YESNOCANCEL = #03      -- Message box contains Yes, No, and Cancel

-- possible values returned by MessageBox() 
-- 0 means failure
global constant IDABORT = 3,  -- Abort button was selected.
                IDCANCEL = 2, -- Cancel button was selected.
                IDIGNORE = 5, -- Ignore button was selected.
                IDNO = 7,     -- No button was selected.
                IDOK = 1,     -- OK button was selected.
                IDRETRY = 4,  -- Retry button was selected.
                IDYES = 6    -- Yes button was selected.

-- the three buttons in the message box
integer
    button1, 
    button2,
    button3

-----------------------------------------------------------------------------
-- icons for the message box

constant
IconInformation = textToBitmap( {
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxx      xxxxxxxx",
    "x   x    xxxxxxx          xxxxxx",
    "x   x    xxxxxx            xxxxx",
    "x        xxxx                xxx",
    "x        xxxx                xxx",
    "x        xxx                  xx",
    "x        xxx   xxx xxx xxx    xx",
    "x        xxx                  xx",
    "x        xxx                  xx",
    "x        xxx   xxx xxx x x    xx",
    "x        xxx                  xx",
    "x        xxx                  xx",
    "x        xxx   xxx x xxx      xx",
    "x        xxx                  xx",
    "x    xxxxxxx                  xx",
    "x      xxxxx   xxxx xxx xx    xx",
    "x      xxxxx                 xxx",
    "x      xxxxx                 xxx",
    "x      xxxxx               xxxxx",
    "x   xxxxx                xxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"} ),


IconQuestion = textToBitmap( {
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxxxxxxxxxxxxxxxx",
    "x        xxxxxxxxx      xxxxxxxx",
    "x   x    xxxxxxx          xxxxxx",
    "x   x    xxxxxx    xxxxx   xxxxx",
    "x        xxxx    xxxxxxxxx   xxx",
    "x        xxxx   xxxxxxxxxxx  xxx",
    "x        xxx    xxxx  xxxxx   xx",
    "x        xxx   xxxx    xxxx   xx",
    "x        xxx   xxxx    xxxx   xx",
    "x        xxx          xxxx    xx",
    "x        xxx         xxxx     xx",
    "x        xxx        xxxx      xx",
    "x        xxx        xxxx      xx",
    "x        xxx       xxxx       xx",
    "x        xxx       xxxx       xx",
    "x    xxxxxxx       xxxx       xx",
    "x      xxxxx                  xx",
    "x      xxxxx       xxxx      xxx",
    "x      xxxxx       xxxx      xxx",
    "x      xxxxx       xxxx    xxxxx",
    "x   xxxxx                xxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "x      xxxxxxxxxxxxxxxxxxxxxxxxx",
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"} ),



IconExclamation = textToBitmap( {
    "               xx               ",
    "              xxxx              ",
    "              xxxx              ",
    "             xx  xx             ",
    "             xx  xx             ",
    "            xx    xx            ",
    "            xx    xx            ",
    "           xx  xx  xx           ",
    "          xx  xxxx  xx          ",
    "          xx  xxxx  xx          ",
    "         xx   xxxx   xx         ",
    "         xx   xxxx   xx         ",
    "        xx    xxxx    xx        ",
    "        xx    xxxx    xx        ",
    "       xx     xxxx     xx       ",
    "       xx     xxxx     xx       ",
    "      xx      xxxx      xx      ",
    "      xx      xxxx      xx      ",
    "     xx       xxxx       xx     ",
    "     xx        xx        xx     ",
    "    xx         xx         xx    ",
    "    xx                    xx    ",
    "   xx                      xx   ",
    "   xx          xx          xx   ",
    "  xx          xxxx          xx  ",
    " xx            xx            xx ",
    " xx                          xx ",
    "xx                            xx",
    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    " xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx "} ),

IconHand = textToBitmap( {
    "                                ",
    "                                ",
    "        xxxxxxxxxxxxxxxx        ",
    "       xxxxxxxxxxxxxxxxxx       ",
    "      xxxxxxxxxxxxxxxxxxxx      ",
    "     xxxxxxxxxx xxxxxxxxxxx     ",
    "    xxxxxxxxxx   xxxxxxxxxxx    ",
    "   xxxxxxxx xx   xx  xxxxxxxx   ",
    "  xxxxxxxx   x   x   xxxxxxxxx  ",
    " xxxxxxxxx   x   x   xxxxxxxxxx ",
    "xxxxxxxx x   x   x   xxxxxxxxxxx",
    "xxxxxxx  x   x   x   xxxxxxxxxxx",
    "xxxxxxx  x   x   x   xxxxxxxxxxx",
    "xxxxxxx  x   x   x   xxxxxxxxxxx",
    "xxxxxxx  x   x   x   xxxxxxxxxxx",
    "xxxxxxx  x   x   x   xxxxxxxxxxx",
    "xxxxxxx  x           xxx   xxxxx",
    "xxxxxxx              xx    xxxxx",
    "xxxxxxx              xx    xxxxx",
    "xxxxxxx              x    xxxxxx",
    "xxxxxxx                   xxxxxx",
    "xxxxxxx                  xxxxxxx",
    "xxxxxxx                 xxxxxxxx",
    "xxxxxxx                 xxxxxxxx",
    "xxxxxxx                xxxxxxxxx",
    "xxxxxxx                xxxxxxxxx",
    " xxxxxx      x        xxxxxxxxx ",
    "  xxxxxx      x      xxxxxxxxx  ",
    "   xxxxxxxxxxxxxxxxxxxxxxxxxx   ",
    "    xxxxxxxxxxxxxxxxxxxxxxxx    ",
    "     xxxxxxxxxxxxxxxxxxxxxx     ",
    "      xxxxxxxxxxxxxxxxxxxx      ",
    "       xxxxxxxxxxxxxxxxxx       "} ),

IconBomb = textToBitmap( {          
    "                    x       x  ",
    "                           x   ",
    "                     x    x    ",
    "                x        x     ",
    "                  x   x        ",
    "                    x          ",
    "           xxxxx          x x x",
    "          x     xx   xxx       ",
    "         x        xxx          ",
    "         x                x    ",
    "         x          x  x   x   ",
    "      xxxxxxx      x        x  ",
    "      xxxxxxx     x    x     x ",
    "      xxxxxxx                  ",
    "    xxxxxxxxxxx        x       ",
    "  xxxxxxxxxxxxxxx              ",
    "  xxxxxxxxxxxxxxx              ",
    " xxxxxxxxxxxxxxxxx             ",
    " xxxxxxxxxxxxxxxxx             ",
    "xxxxxxxxxxxxxx xxxx            ",
    "xxxxxxxxxxxxxxxxxxx            ",
    "xxxxxxxxxxxxxxx xxx            ",
    "xxxxxxxxxxxxxxx xxx            ",
    "xxxxxxxxxxxxxxx xxx            ",
    "xxxxxxxxxxxxxxxxxxx            ",
    "xxxxxxxxxxxxxx xxxx            ",
    " xxxxxxxxxxxxxxxxx             ",
    " xxxxxxxxxxx xxxxx             ",
    "  xxxxxxxxxxxxxxx              ",
    "  xxxxxxxxxxxxxxx              ",
    "    xxxxxxxxxxx                ",
    "      xxxxxxx                  "})


constant
-- the message box and controls
    MsgBox       = create( Window, "",  0, Default, Default, 200, 100, 0 ),
    MBoxButton1  = create( PushButton, "", MsgBox, 20, 50, 80, 20, 0 ),
    MBoxButton2  = create( PushButton, "", MsgBox, 100, 50, 80, 20, 0 ),
    MBoxButton3  = create( PushButton, "", MsgBox, 100, 50, 80, 20, 0 ),
    MBoxText     = create( LText, "", MsgBox, 100, 100, 100, 20, 0 ),
    MBoxIcon     = create( Bitmap, "", MsgBox, 10, 10, 32, 32, 0 )

integer MBoxResult
    MBoxResult = 0
    
-----------------------------------------------------------------------------
procedure msgBoxButton1()
    -- set result and close window
    MBoxResult = button1
    closeWindow( MsgBox )    
end procedure

-----------------------------------------------------------------------------
procedure msgBoxButton2()
    -- set result and close window    
    MBoxResult = button2
    closeWindow( MsgBox )    
end procedure

-----------------------------------------------------------------------------
procedure msgBoxButton3()
    -- set result and close window    
    MBoxResult = button3
    closeWindow( MsgBox )    
end procedure


-- assign behaviors
onClick[MBoxButton1]    = routine_id("msgBoxButton1")
onClick[MBoxButton2]    = routine_id("msgBoxButton2")
onClick[MBoxButton3]    = routine_id("msgBoxButton2")


-----------------------------------------------------------------------------
function bits( atom a, atom bits )
    -- true if all bits in pattern are set
    return and_bits( a, bits ) = bits
end function


-----------------------------------------------------------------------------
global function message_box( sequence text, sequence title, integer flags )
    -- $ Euphoria Win32 Emulation
    -- Emulation of Euphoria message_box function
    integer message, x1, y1, x2, y2
    sequence size

    -- hide the mouse
    mouse_pointer(0)

    -- save the flags
    my[MsgBox][Flags] = flags

    -----------
    -- the icon
    -----------

    -- set the icon. the bits are not exclusive, so i have to
    -- test that it matches *all* the bits of the flag

    if flags = -1 then
        -- this is *my* special code
        setBitmap( MBoxIcon, IconBomb)

    -- this uses more than 1 bit
    elsif bits( flags, MB_ICONEXCLAMATION ) then
        setBitmap( MBoxIcon, IconExclamation )
        
    elsif bits( flags, MB_ICONINFORMATION ) then
        setBitmap( MBoxIcon, IconInformation )
    
    elsif bits( flags, MB_ICONERROR ) then
        setBitmap( MBoxIcon, IconHand )
        
    elsif bits( flags, MB_ICONQUESTION ) then
        setBitmap( MBoxIcon, IconQuestion )
        
    else
        setBitmap( MBoxIcon, 0 )
        
    end if

    --------------------
    -- position the text
    --------------------

    -- get the size of the text
    size = getTextRect( text, 0 )

    -- position the text
    if my[MBoxIcon][Bmp] then     
        -- adjust for the icon        
        size = size + { 60, 20, 60, 20 }
    else
        -- no icon
        size = size + { 10, 10, 10, 10 }
    end if   
    
    -- set the size
    my[MBoxText][Size] = size

    -- set the text
    setText( MsgBox, title )
    setText( MBoxText, text )
    
    -- get the size for positioning various other elements
    x1 = size[X1]
    y1 = size[Y1]
    x2 = size[X2]
    y2 = size[Y2]

    --------------
    -- the buttons
    --------------

    -- get low 3 bits on flags - button codes
    flags = and_bits( flags, 7 )

    -- #01
    if flags = MB_OKCANCEL then -- #01
        -- text
        setText( MBoxButton1, "Ok" )
        setText( MBoxButton2, "Retry" )
        
        -- values
        button1 = IDOK
        button2 = IDCANCEL
        button3 = 0

    -- #02    
    elsif flags = MB_ABORTRETRYIGNORE then
        -- text
        setText( MBoxButton1, "Abort" )
        setText( MBoxButton2, "Retry" )
        setText( MBoxButton3, "Ignore" )
        
        -- values
        button1 = IDABORT
        button2 = IDRETRY
        button3 = IDIGNORE
    
    -- #03
    elsif flags = MB_YESNOCANCEL then
        -- text
        setText( MBoxButton1, "Yes" )
        setText( MBoxButton2, "No" )
        setText( MBoxButton3, "Cancel" )
        
        -- values
        button1 = IDYES
        button2 = IDNO
        button3 = IDCANCEL

    -- #04
    elsif flags = MB_YESNO then
        -- text
        setText( MBoxButton1, "Yes" )
        setText( MBoxButton2, "No" )
        
        -- values
        button1 = IDYES
        button2 = IDNO
        button3 = 0
    
    -- #05
    elsif flags = MB_RETRYCANCEL then
        -- text
        setText( MBoxButton1, "Retry" )
        setText( MBoxButton2, "Cancel" )        
    
        -- values        
        button1 = IDRETRY
        button2 = IDCANCEL
        button3 = 0
        
    else 
        -- text
        setText( MBoxButton1, "Ok" )
        
        -- values
        button1 = IDOK
        button2 = 0
        button3 = 0
    
    end if        

    -- activate appropriate buttons        
    setVisible( MBoxButton1, button1 )
    setVisible( MBoxButton2, button2 )
    setVisible( MBoxButton3, button3 )

    -- position buttons relative to text
    my[MBoxButton1][Size] = { 20, 20+y2, 80, 20 }
    my[MBoxButton2][Size] = {120, 20+y2, 80, 20 }
    my[MBoxButton3][Size] = {220, 20+y2, 80, 20 }


    -----------------
    -- window metrics
    -----------------

    -- add the title bar, border and buttons
    x2 = x2 + (theBorderSize*2)
    y2 = y2 + theTitleSize + (theBorderSize*2) + 46
                       
    -- minimum height
    y2 = max( y2, 100 )                    
    
    -- number of buttons determine minimum width
    if button3 then     x2 = max( x2, 330 )
    elsif button2 then  x2 = max( x2, 220 )
    else                x2 = max( x2, 120 )
    end if

    -- position box
    x1 = floor( (xPixels - x2) /2 )
    y1 = floor( (yPixels - y2) /2 )
    my[MsgBox][Size] = { x1, y1, x2, y2 }


    ----------------
    -- set up to run
    ----------------
    

    -- make it modal
    my[MsgBox][Modal] = 1
    
    -- call the main loop
    eventLoop( MsgBox )
    
    -- show the mouse
    mouse_pointer(1)
    
    -- return result
    return MBoxResult
    
end function



-----------------------------------------------------------------------------
procedure abortErr2( sequence error, sequence args )

    -- More complex error handler; this uses windows
    integer result

    -- pop up message box
    result = message_box( sprintf( error, args ), "Program Aborting",
                MB_ICONEXCLAMATION )
    
    -- dos message
    abortErr1( error, args )
    
end procedure

abortHook = routine_id("abortErr2")


-----------------------------------------------------------------------------
-- Main Routine

-----------------------------------------------------------------------------
global procedure WinMain( integer mainWin )

    -- set the main window, and start the event loop
    integer result

    -- is it a window?    
    if my[mainWin][Class] != Window then
        abortErr( "WinMain passed a non-window parameter.", {} )        
    end if

    -- main loop
    eventLoop( mainWin )

    -- restore mode
    abortErr1( "Goodbye from Dos32Lib.", {} )
    

end procedure


-- In Euphoria 2.1 this file (unless modified) is FREE - 0 statements.
-- with 9807020441 -- delete this statement if it causes an error

