-- Object-Oriented Euphoria Library v0.1
-- March 29, 1999
-- By Matthew Whitehead

with trace

without warning

global integer ClassError

ClassError = 0

global constant FALSE = 0

constant TRUE = 1

-- Class constants
global constant CLSNAME = 1,
    PROPS = 2,
    CLSCREATE = 3,
    CLSDESTROY = 4,
    METHODS = 5,
    PARENT = 6
    
-- Object constants
global constant OBJCLASS = 1,
    OBJNAME = 2,
    OBJPROPS = 3

-- Property constants 
constant PROPNAME = 1,
    PROPDEF = 2,
    PROPGET = 3,
    PROPLET = 4
                
-- Method constants
global constant MNAME = 1,
    METHODID = 2

global sequence classes
global sequence objects
global sequence evtHandlers
global sequence currentObj

classes = {}
objects = {}
evtHandlers = {}
currentObj = {}

type boolean(integer x)
    return x = 1 or x = 0    
end type
----

global procedure ErrorMessage(sequence msg)
    clear_screen()
    puts(1, "Object-Oriented Euphoria Error:\n" & msg)
    abort(1)
end procedure

-- Begin find functions

global function findMethod(sequence class, sequence method)
    integer findPos
    integer methPos

    findPos = 0

    for a = 1 to length(classes) do
        if compare(classes[a][CLSNAME], class) = 0 then
            findPos = a
        end if
    end for
    
    if findPos = 0 then
        ErrorMessage(class & " is not a valid class name.")
    end if

    for a = 1 to length(classes[findPos][METHODS]) do
        if compare(classes[findPos][METHODS][a][MNAME], method) = 0 then
            methPos = a
            return methPos
        end if
    end for
    
    ErrorMessage(currentObj[OBJNAME] & "." & method & " is not a valid method.")

    return 0

end function

global function findProp(sequence class, sequence prop)
    integer findPos
    integer propPos

    findPos = 0

    for a = 1 to length(classes) do
        if compare(classes[a][CLSNAME], class) = 0 then
            findPos = a
        end if
    end for
    
    if findPos = 0 then
        ErrorMessage(class & " is not a valid class name.")
    end if

    for a = 1 to length(classes[findPos][PROPS]) do
        if compare(classes[findPos][PROPS][a][1], prop) = 0 then
            propPos = a
            return propPos
            exit
        end if
    end for
    
    if propPos = 0 then
        ErrorMessage(prop & " is not a valid property for class " & class & ".")
    end if

    return 0

end function

global function findObj(sequence name)
    integer findPos

    findPos = 0

    for a = 1 to length(objects) do
        if compare(objects[a][OBJNAME], name) = 0 then
            findPos = a
            return findPos
            exit
        end if
    end for
        
    ErrorMessage(name & " is not a valid object name.")

end function

global function findClass(sequence class)
    integer findPos

    findPos = 0

    for a = 1 to length(classes) do
        if compare(classes[a][CLSNAME], class) = 0 then
            findPos = a
            return findPos
            exit
        end if
    end for
    
    if ClassError = 0 then
        ErrorMessage(class & " has not been registered.")
    else
        ClassError = 2
        return 0
    end if
    
end function

global procedure ClassExist(sequence class)
    integer pos

    pos = findClass(class)

    if pos = 0 then
        ErrorMessage(class & " has not been registered.")
    end if

end procedure

-- See if object supports interface clsName
global function QueryInterface(sequence objName, sequence clsName)
    integer objPos, clsPos
    sequence cls, obj
    
    objPos = findObj(objName)
    clsPos = findClass(clsName)
    
    obj = objects[objPos][OBJCLASS]
        
    if equal(obj, clsName) then
        return TRUE
    else
        cls = objects[objPos][OBJCLASS]    
        clsPos = findClass(cls)
        cls = classes[clsPos][PARENT]
        
        while 1 do                                        
            if equal(cls, clsName) then
                return TRUE
            else        
                if length(classes[clsPos][PARENT]) = 0 then
                    return FALSE    
                end if
                clsPos = findClass(classes[clsPos][PARENT])
                cls = classes[clsPos][CLSNAME]
            end if
        end while
    end if      
    
    return FALSE
    
end function                            


-- Get class of object

global function TypeName(sequence objName)
    integer objPos

    objPos = findObj(objName)

    return objects[objPos][OBJCLASS]

end function

global procedure ChangeProp(sequence objName, sequence prop, object value)
    integer propPos, objPos
    sequence objProps, obj
    
    objPos = findObj(objName)

    obj = objects[objPos]

    objProps = obj[OBJPROPS]
    
    propPos = findProp(obj[OBJCLASS], prop)

    if propPos = -1 then
        ErrorMessage(prop & " is not a valid property for class " & obj[OBJCLASS] & ".")    
    end if
    
    objProps[propPos] = value
    
    objects[objPos][OBJPROPS] = objProps

end procedure

-- Begin registration procedures

global procedure RegisterEvent(sequence objName, sequence event, integer evtProc)
    sequence evt

    evt = {objName, event, evtProc}
    
    evtHandlers = append(evtHandlers, evt)
    
end procedure

global procedure RegisterProp(sequence class, sequence prop, integer GetCallback, 
    integer LetCallback, object defValue)

    integer findPos
 
    findPos = findClass(class)

    classes[findPos][PROPS] = append(classes[findPos][PROPS], {prop, defValue, GetCallback, LetCallback})
 
end procedure

global procedure RegisterMethod(sequence class, sequence method, integer callback)
    integer pos

    pos = findClass(class)

    classes[pos][METHODS] = append(classes[pos][METHODS], {method, callback})

end procedure

procedure extends(sequence extendclass, sequence class, integer CreateProc, integer DestroyProc)
    integer pos, extPos
    sequence cls, extCls

    pos = findClass(class)
    cls = classes[pos]

    extPos = findClass(extendclass)
    extCls = classes[extPos]
    
    for a = 1 to length(extCls[PROPS]) do
        RegisterProp(class, extCls[PROPS][a][PROPNAME], extCls[PROPS][a][PROPGET], extCls[PROPS][a][PROPLET], extCls[PROPS][a][PROPDEF])
    end for

    for a = 1 to length(extCls[METHODS]) do
        RegisterMethod(class, extCls[METHODS][a][MNAME], extCls[METHODS][a][METHODID])
    end for
    
    if CreateProc > -1 then
        cls[CLSCREATE] = CreateProc
    else
        cls[CLSCREATE] = extCls[CLSCREATE]
    end if                         
    
    if DestroyProc > -1 then
        cls[CLSDESTROY] = DestroyProc     
    else
        cls[CLSDESTROY] = extCls[CLSCREATE]   
    end if                           
    
end procedure

global procedure OverrideMethod(sequence class, sequence method, integer NewCallback)

    integer clsPos, methPos

    clsPos = findClass(class)
    methPos = findMethod(class, method)
    
    classes[clsPos][METHODS][methPos][METHODID] = NewCallback

end procedure

global procedure OverrideProp(sequence class, sequence prop, integer newGet, integer newLet)
    integer clsPos, propPos
    
    clsPos = findClass(class)
    propPos = findProp(class, prop)
    
    classes[clsPos][PROPS][propPos][PROPGET] = newGet
    classes[clsPos][PROPS][propPos][PROPLET] = newLet
            
end procedure

global procedure RegisterClass(sequence class, integer createProc, integer DestroyProc, sequence superCls)

    classes = append(classes, {class, {}, createProc, DestroyProc, {}, superCls})
    
    if length(superCls) > 0then
        extends (superCls, class, createProc, DestroyProc)    
    end if
    
end procedure


global procedure setObject(sequence obj1, sequence obj2)
    integer obj1Pos, obj2Pos
    integer cls1Pos, cls2Pos
    
    boolean objMatch

    obj1Pos = findObj(obj1)
    obj2Pos = findObj(obj2)
    
    -- Make sure obj2 supports interface of obj1
    objMatch = QueryInterface(obj2, objects[obj1Pos][CLSNAME])

    if objMatch = FALSE then
        ErrorMessage(obj1 & " is not a superclass of " & obj2 & ".")        
    end if                                                        
    
    for a = 1 to length(objects[obj1Pos][OBJPROPS]) do
        objects[obj1Pos][OBJPROPS][a] = objects[obj2Pos][OBJPROPS][a]                
    end for
    
    cls1Pos = findClass(objects[obj1Pos][OBJCLASS])
    cls2Pos = findClass(objects[obj2Pos][OBJCLASS])
    
    for a = 1 to length(classes[cls1Pos][METHODS]) do
        classes[cls1Pos][METHODS][a][METHODID] = classes[cls2Pos][METHODS][a][METHODID]                    
    end for
    
end procedure                                    

-- End registration routines

-- Event Handling Procedures

function FindEventHandler(sequence objName, sequence event)
    integer callback
    sequence evtHndl

    for a = 1 to length(evtHandlers) do
        evtHndl = evtHandlers[a]
        if compare(objName, evtHndl[1]) = 0 and
             compare(event, evtHndl[2]) = 0 then
             return evtHndl[3]
        end if
    end for
    
    return -1

end function

global procedure RaiseEvent(sequence obj, sequence event, sequence args)
    integer eventID
    
    eventID = FindEventHandler(obj, event)
    
    if eventID > -1 then
            call_proc(eventID, args)
    end if
    
end procedure

-- Object routines

global procedure new(sequence class, sequence name, sequence args)
    integer createProc
    integer classPos
    integer retVal
    sequence seqObject

    -- Object Definition
    -- 
    -- Each object has a class, registered with RegisterClass, and the properties of that class,
    -- registered with RegisterProp.
    -- All objects are stored in the objects sequence.
    -- 
    -- objects = {{"MyClass", "ObjName", {"Caption", 1, ...}}, {"AnotherClass", "Obj2", {"Properties", 2, ...}}}
    
    seqObject = {}
    
    classPos = findClass(class)
    
    if classPos > 0 then
            
        createProc = classes[classPos][CLSCREATE]

        for a = 1 to length(classes[classPos][PROPS]) do
            seqObject = append(seqObject, classes[classPos][PROPS][a][PROPDEF])
        end for                                                               

        objects = append(objects, {class, name, seqObject})

        if createProc = -1 then
            -- ErrorMessage(class & " doesn't have a Create function.")    
        else
            args = prepend(args, name)

            retVal = call_func(createProc, args)

            if retVal = 0 then
                ErrorMessage(class & " Create failed.")
            end if
        
        end if
    
    end if
    
end procedure

global procedure destroy(sequence name)
    integer pos
    integer clsPos

    pos = findObj(name)

    if pos = 0 then
        ErrorMessage(name & " is not a valid object name.")
    end if

    clsPos = findClass(objects[pos][OBJCLASS])

    if classes[clsPos][CLSDESTROY] > -1 then
        call_proc(classes[clsPos][CLSDESTROY], {name})
    end if

    if length(objects) = pos then
        objects = objects[1..pos-1]
    else
        objects = objects[1..pos-1] & objects[pos+1..length(objects)]
    end if

end procedure

-- Properties procedures

global procedure setProp(sequence name, sequence prop, object NewValue)
    integer propProc
    integer objPos
    integer clsPos
    integer propPos
    sequence retVal
    
    objPos = findObj(name)

    if objPos = 0 then
        ErrorMessage(name & " is not a valid object name.")
    end if

    clsPos = findClass(objects[objPos][OBJCLASS])

    propPos = findProp(objects[objPos][OBJCLASS], prop)
 
    propProc = classes[clsPos][PROPS][propPos][4]
    
    currentObj = objects[objPos]
    
    if propProc = -1 then
        ErrorMessage("Property " & prop & " is a read-only property in class " & objects[objPos][1] & ".")
    elsif propProc = -2 then
        ChangeProp(name, prop, NewValue)
    else
        retVal = call_func(propProc, {objects[objPos][OBJNAME], NewValue})
        if retVal[1] = 1 then
            RaiseEvent(name, retVal[2], retVal[3])    
        end if
        ChangeProp(name, prop, NewValue)
    end if
    
    currentObj = {}
    
end procedure

global function getProp(sequence name, sequence prop)
    integer propProc
    integer objPos
    integer propPos
    integer clsPos

    object propValue

    objPos = findObj(name)

    if objPos = 0 then
        ErrorMessage(name & " is not a valid object name.")
    end if

    propPos = findProp(objects[objPos][OBJCLASS], prop)
 
    if propPos = 0 then
        ErrorMessage(prop & " is not a valid property for class " & objects[objPos][OBJCLASS] & ".")
    end if

    clsPos = findClass(objects[objPos][1])

    propProc = classes[clsPos][PROPS][propPos][3]
    
    currentObj = objects[objPos]
    
    if propProc = -1 then
        ErrorMessage("Property " & prop & " is a private property in class " & objects[objPos][OBJCLASS] & ".")
    elsif propProc = -2 then
        propValue = objects[objPos][OBJPROPS][propPos]
    else 
        propValue = call_func(propProc, {objects[objPos][OBJNAME], objects[objPos]})    
    end if
    
    currentObj = {}
    
    return propValue

end function


-- Method & Function Calling Routines

global procedure method(sequence name, sequence method, sequence args)
    integer objPos
    integer clsPos
    integer methPos
    integer methProc

    objPos = findObj(name)

    clsPos = findClass(objects[objPos][OBJCLASS])

    methPos = findMethod(objects[objPos][OBJCLASS], method)

    methProc = classes[clsPos][METHODS][methPos][METHODID]

    if methProc = -1 then
        ErrorMessage("Class " & objects[objPos][OBJCLASS] & " needs to implement method " &
            method & ".")
    end if

    if length(args) = 0 then
        args = objects[objPos][OBJNAME]
        call_proc(methProc, {args})
    else
        args = prepend(args, objects[objPos][OBJNAME])
        call_proc(methProc, args)
    end if

end procedure

global function func(sequence name, sequence funcname, sequence args)
    integer objPos
    integer clsPos
    integer methPos
    integer methProc
    object retValue

    objPos = findObj(name)

    clsPos = findClass(objects[objPos][OBJCLASS])

    methPos = findMethod(objects[objPos][OBJCLASS], funcname)

    methProc = classes[clsPos][METHODS][methPos][METHODID]

    args = prepend(args, objects[objPos][OBJNAME])
    
    if length(args) = 1 then
        retValue = call_func(methProc, {args})
    else
        retValue = call_func(methProc, args)
    end if

    return retValue

end function

-- END of ObjEu.e

